"""Unit/functional testing for argparse completer in cmd2"""

import argparse
import numbers
from typing import cast

import pytest
from rich.text import Text

import cmd2
import cmd2.string_utils as su
from cmd2 import (
    Cmd2ArgumentParser,
    CompletionError,
    CompletionItem,
    argparse_completer,
    argparse_custom,
    with_argparser,
)
from cmd2 import rich_utils as ru

from .conftest import (
    complete_tester,
    normalize,
    run_cmd,
    with_ansi_style,
)

# Data and functions for testing standalone choice_provider and completer
standalone_choices = ['standalone', 'provider']
standalone_completions = ['standalone', 'completer']


def standalone_choice_provider(cli: cmd2.Cmd) -> list[str]:
    return standalone_choices


def standalone_completer(cli: cmd2.Cmd, text: str, line: str, begidx: int, endidx: int) -> list[str]:
    return cli.basic_complete(text, line, begidx, endidx, standalone_completions)


class ArgparseCompleterTester(cmd2.Cmd):
    """Cmd2 app that exercises ArgparseCompleter class"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    ############################################################################################################
    # Begin code related to help and command name completion
    ############################################################################################################
    # Top level parser for music command
    music_parser = Cmd2ArgumentParser(description='Manage music')

    # Add subcommands to music
    music_subparsers = music_parser.add_subparsers()
    music_create_parser = music_subparsers.add_parser('create', help='create music')

    # Add subcommands to music -> create
    music_create_subparsers = music_create_parser.add_subparsers()
    music_create_jazz_parser = music_create_subparsers.add_parser('jazz', help='create jazz')
    music_create_rock_parser = music_create_subparsers.add_parser('rock', help='create rocks')

    @with_argparser(music_parser)
    def do_music(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to flag completion
    ############################################################################################################

    # Uses default flag prefix value (-)
    flag_parser = Cmd2ArgumentParser()
    flag_parser.add_argument('-n', '--normal_flag', help='a normal flag', action='store_true')
    flag_parser.add_argument('-a', '--append_flag', help='append flag', action='append')
    flag_parser.add_argument('-o', '--append_const_flag', help='append const flag', action='append_const', const=True)
    flag_parser.add_argument('-c', '--count_flag', help='count flag', action='count')
    flag_parser.add_argument('-s', '--suppressed_flag', help=argparse.SUPPRESS, action='store_true')
    flag_parser.add_argument('-r', '--remainder_flag', nargs=argparse.REMAINDER, help='a remainder flag')
    flag_parser.add_argument('-q', '--required_flag', required=True, help='a required flag', action='store_true')

    @with_argparser(flag_parser)
    def do_flag(self, args: argparse.Namespace) -> None:
        pass

    # Uses non-default flag prefix value (+)
    plus_flag_parser = Cmd2ArgumentParser(prefix_chars='+')
    plus_flag_parser.add_argument('+n', '++normal_flag', help='a normal flag', action='store_true')
    plus_flag_parser.add_argument('+q', '++required_flag', required=True, help='a required flag', action='store_true')

    @with_argparser(plus_flag_parser)
    def do_plus_flag(self, args: argparse.Namespace) -> None:
        pass

    # A parser with a positional and flags. Used to test that remaining flag names are completed when all positionals are done.
    pos_and_flag_parser = Cmd2ArgumentParser()
    pos_and_flag_parser.add_argument("positional", choices=["a", "choice"])
    pos_and_flag_parser.add_argument("-f", "--flag", action='store_true')

    @with_argparser(pos_and_flag_parser)
    def do_pos_and_flag(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to testing choices and choices_provider parameters
    ############################################################################################################
    STR_METAVAR = "HEADLESS"
    TUPLE_METAVAR = ('arg1', 'others')
    CUSTOM_DESC_HEADERS = ("Custom Headers",)

    # tuples (for sake of immutability) used in our tests (there is a mix of sorted and unsorted on purpose)
    non_negative_num_choices = (1, 2, 3, 0.5, 22)
    num_choices = (-1, 1, -2, 2.5, 0, -12)
    static_choices_list = ('static', 'choices', 'stop', 'here')
    choices_from_provider = ('choices', 'provider', 'probably', 'improved')
    completion_item_choices = (
        CompletionItem('choice_1', ['Description 1']),
        # Make this the longest description so we can test display width.
        CompletionItem('choice_2', [su.stylize("String with style", style=cmd2.Color.BLUE)]),
        CompletionItem('choice_3', [Text("Text with style", style=cmd2.Color.RED)]),
    )

    # This tests that CompletionItems created with numerical values are sorted as numbers.
    num_completion_items = (
        CompletionItem(5, ["Five"]),
        CompletionItem(1.5, ["One.Five"]),
        CompletionItem(2, ["Five"]),
    )

    def choices_provider(self) -> tuple[str]:
        """Method that provides choices"""
        return self.choices_from_provider

    def completion_item_method(self) -> list[CompletionItem]:
        """Choices method that returns CompletionItems"""
        items = []
        for i in range(10):
            main_str = f'main_str{i}'
            items.append(CompletionItem(main_str, ['blah blah']))
        return items

    choices_parser = Cmd2ArgumentParser()

    # Flag args for choices command. Include string and non-string arg types.
    choices_parser.add_argument("-l", "--list", help="a flag populated with a choices list", choices=static_choices_list)
    choices_parser.add_argument(
        "-p", "--provider", help="a flag populated with a choices provider", choices_provider=choices_provider
    )
    choices_parser.add_argument(
        "--desc_header",
        help='this arg has a descriptive header',
        choices_provider=completion_item_method,
        descriptive_headers=CUSTOM_DESC_HEADERS,
    )
    choices_parser.add_argument(
        "--no_header",
        help='this arg has no descriptive header',
        choices_provider=completion_item_method,
        metavar=STR_METAVAR,
    )
    choices_parser.add_argument(
        '-t',
        "--tuple_metavar",
        help='this arg has tuple for a metavar',
        choices_provider=completion_item_method,
        metavar=TUPLE_METAVAR,
        nargs=argparse.ONE_OR_MORE,
    )
    choices_parser.add_argument('-n', '--num', type=int, help='a flag with an int type', choices=num_choices)
    choices_parser.add_argument('--completion_items', help='choices are CompletionItems', choices=completion_item_choices)
    choices_parser.add_argument(
        '--num_completion_items', help='choices are numerical CompletionItems', choices=num_completion_items
    )

    # Positional args for choices command
    choices_parser.add_argument("list_pos", help="a positional populated with a choices list", choices=static_choices_list)
    choices_parser.add_argument(
        "method_pos", help="a positional populated with a choices provider", choices_provider=choices_provider
    )
    choices_parser.add_argument(
        'non_negative_num', type=int, help='a positional with non-negative numerical choices', choices=non_negative_num_choices
    )
    choices_parser.add_argument('empty_choices', help='a positional with empty choices', choices=[])

    @with_argparser(choices_parser)
    def do_choices(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to testing completer parameter
    ############################################################################################################
    completions_for_flag = ('completions', 'flag', 'fairly', 'complete')
    completions_for_pos_1 = ('completions', 'positional_1', 'probably', 'missed', 'spot')
    completions_for_pos_2 = ('completions', 'positional_2', 'probably', 'missed', 'me')

    def flag_completer(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self.basic_complete(text, line, begidx, endidx, self.completions_for_flag)

    def pos_1_completer(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self.basic_complete(text, line, begidx, endidx, self.completions_for_pos_1)

    def pos_2_completer(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self.basic_complete(text, line, begidx, endidx, self.completions_for_pos_2)

    completer_parser = Cmd2ArgumentParser()

    # Flag args for completer command
    completer_parser.add_argument("-c", "--completer", help="a flag using a completer", completer=flag_completer)

    # Positional args for completer command
    completer_parser.add_argument("pos_1", help="a positional using a completer method", completer=pos_1_completer)
    completer_parser.add_argument("pos_2", help="a positional using a completer method", completer=pos_2_completer)

    @with_argparser(completer_parser)
    def do_completer(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to nargs
    ############################################################################################################
    set_value_choices = ('set', 'value', 'choices')
    one_or_more_choices = ('one', 'or', 'more', 'choices')
    optional_choices = ('a', 'few', 'optional', 'choices')
    range_choices = ('some', 'range', 'choices')
    remainder_choices = ('remainder', 'choices')
    positional_choices = ('the', 'positional', 'choices')

    nargs_parser = Cmd2ArgumentParser()

    # Flag args for nargs command
    nargs_parser.add_argument("--set_value", help="a flag with a set value for nargs", nargs=2, choices=set_value_choices)
    nargs_parser.add_argument(
        "--one_or_more", help="a flag wanting one or more args", nargs=argparse.ONE_OR_MORE, choices=one_or_more_choices
    )
    nargs_parser.add_argument(
        "--optional", help="a flag with an optional value", nargs=argparse.OPTIONAL, choices=optional_choices
    )
    nargs_parser.add_argument("--range", help="a flag with nargs range", nargs=(1, 2), choices=range_choices)
    nargs_parser.add_argument(
        "--remainder", help="a flag wanting remaining", nargs=argparse.REMAINDER, choices=remainder_choices
    )

    nargs_parser.add_argument("normal_pos", help="a remainder positional", nargs=2, choices=positional_choices)
    nargs_parser.add_argument(
        "remainder_pos", help="a remainder positional", nargs=argparse.REMAINDER, choices=remainder_choices
    )

    @with_argparser(nargs_parser)
    def do_nargs(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to testing tab hints
    ############################################################################################################
    hint_parser = Cmd2ArgumentParser()
    hint_parser.add_argument('-f', '--flag', help='a flag arg')
    hint_parser.add_argument('-s', '--suppressed_help', help=argparse.SUPPRESS)
    hint_parser.add_argument('-t', '--suppressed_hint', help='a flag arg', suppress_tab_hint=True)

    hint_parser.add_argument('hint_pos', help='here is a hint\nwith new lines')
    hint_parser.add_argument('no_help_pos')

    @with_argparser(hint_parser)
    def do_hint(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to CompletionError
    ############################################################################################################
    def completer_raise_error(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        """Raises CompletionError"""
        raise CompletionError('completer broke something')

    def choice_raise_error(self) -> list[str]:
        """Raises CompletionError"""
        raise CompletionError('choice broke something')

    comp_error_parser = Cmd2ArgumentParser()
    comp_error_parser.add_argument('completer_pos', help='positional arg', completer=completer_raise_error)
    comp_error_parser.add_argument('--choice', help='flag arg', choices_provider=choice_raise_error)

    @with_argparser(comp_error_parser)
    def do_raise_completion_error(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to receiving arg_tokens
    ############################################################################################################
    def choices_takes_arg_tokens(self, arg_tokens: dict[str, list[str]]) -> list[str]:
        """Choices function that receives arg_tokens from ArgparseCompleter"""
        return [arg_tokens['parent_arg'][0], arg_tokens['subcommand'][0]]

    def completer_takes_arg_tokens(
        self, text: str, line: str, begidx: int, endidx: int, arg_tokens: dict[str, list[str]]
    ) -> list[str]:
        """Completer function that receives arg_tokens from ArgparseCompleter"""
        match_against = [arg_tokens['parent_arg'][0], arg_tokens['subcommand'][0]]
        return self.basic_complete(text, line, begidx, endidx, match_against)

    arg_tokens_parser = Cmd2ArgumentParser()
    arg_tokens_parser.add_argument('parent_arg', help='arg from a parent parser')

    # Create a subcommand for to exercise receiving parent_tokens and subcommand name in arg_tokens
    arg_tokens_subparser = arg_tokens_parser.add_subparsers(dest='subcommand')
    arg_tokens_subcmd_parser = arg_tokens_subparser.add_parser('subcmd')

    arg_tokens_subcmd_parser.add_argument('choices_pos', choices_provider=choices_takes_arg_tokens)
    arg_tokens_subcmd_parser.add_argument('completer_pos', completer=completer_takes_arg_tokens)

    # Used to override parent_arg in arg_tokens_parser
    arg_tokens_subcmd_parser.add_argument('--parent_arg')

    @with_argparser(arg_tokens_parser)
    def do_arg_tokens(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to mutually exclusive groups
    ############################################################################################################
    mutex_parser = Cmd2ArgumentParser()

    mutex_group = mutex_parser.add_mutually_exclusive_group(required=True)
    mutex_group.add_argument('optional_pos', help='the optional positional', nargs=argparse.OPTIONAL)
    mutex_group.add_argument('-f', '--flag', help='the flag arg')
    mutex_group.add_argument('-o', '--other_flag', help='the other flag arg')

    mutex_parser.add_argument('last_arg', help='the last arg')

    @with_argparser(mutex_parser)
    def do_mutex(self, args: argparse.Namespace) -> None:
        pass

    ############################################################################################################
    # Begin code related to standalone functions
    ############################################################################################################
    standalone_parser = Cmd2ArgumentParser()
    standalone_parser.add_argument('--provider', help='standalone provider', choices_provider=standalone_choice_provider)
    standalone_parser.add_argument('--completer', help='standalone completer', completer=standalone_completer)

    @with_argparser(standalone_parser)
    def do_standalone(self, args: argparse.Namespace) -> None:
        pass


@pytest.fixture
def ac_app():
    return ArgparseCompleterTester()


@pytest.mark.parametrize('command', ['music', 'music create', 'music create rock', 'music create jazz'])
def test_help(ac_app, command) -> None:
    out1, _err1 = run_cmd(ac_app, f'{command} -h')
    out2, _err2 = run_cmd(ac_app, f'help {command}')
    assert out1 == out2


def test_bad_subcommand_help(ac_app) -> None:
    # These should give the same output because the second one isn't using a
    # real subcommand, so help will be called on the music command instead.
    out1, _err1 = run_cmd(ac_app, 'help music')
    out2, _err2 = run_cmd(ac_app, 'help music fake')
    assert out1 == out2


@pytest.mark.parametrize(
    ('command', 'text', 'completions'),
    [
        ('', 'mus', ['music ']),
        ('music', 'cre', ['create ']),
        ('music', 'creab', []),
        ('music create', '', ['jazz', 'rock']),
        ('music crea', 'jazz', []),
        ('music create', 'foo', []),
        ('fake create', '', []),
        ('music fake', '', []),
    ],
)
def test_complete_help(ac_app, command, text, completions) -> None:
    line = f'help {command} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('subcommand', 'text', 'completions'),
    [('create', '', ['jazz', 'rock']), ('create', 'ja', ['jazz ']), ('create', 'foo', []), ('creab', 'ja', [])],
)
def test_subcommand_completions(ac_app, subcommand, text, completions) -> None:
    line = f'music {subcommand} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('command_and_args', 'text', 'completion_matches', 'display_matches'),
    [
        # Complete all flags (suppressed will not show)
        (
            'flag',
            '-',
            [
                '--append_const_flag',
                '--append_flag',
                '--count_flag',
                '--help',
                '--normal_flag',
                '--remainder_flag',
                '--required_flag',
                '-a',
                '-c',
                '-h',
                '-n',
                '-o',
                '-q',
                '-r',
            ],
            [
                '-q, --required_flag',
                '[-o, --append_const_flag]',
                '[-a, --append_flag]',
                '[-c, --count_flag]',
                '[-h, --help]',
                '[-n, --normal_flag]',
                '[-r, --remainder_flag]',
            ],
        ),
        (
            'flag',
            '--',
            [
                '--append_const_flag',
                '--append_flag',
                '--count_flag',
                '--help',
                '--normal_flag',
                '--remainder_flag',
                '--required_flag',
            ],
            [
                '--required_flag',
                '[--append_const_flag]',
                '[--append_flag]',
                '[--count_flag]',
                '[--help]',
                '[--normal_flag]',
                '[--remainder_flag]',
            ],
        ),
        # Complete individual flag
        ('flag', '-n', ['-n '], ['[-n]']),
        ('flag', '--n', ['--normal_flag '], ['[--normal_flag]']),
        # No flags should complete until current flag has its args
        ('flag --append_flag', '-', [], []),
        # Complete REMAINDER flag name
        ('flag', '-r', ['-r '], ['[-r]']),
        ('flag', '--rem', ['--remainder_flag '], ['[--remainder_flag]']),
        # No flags after a REMAINDER should complete
        ('flag -r value', '-', [], []),
        ('flag --remainder_flag value', '--', [], []),
        # Suppressed flag should not complete
        ('flag', '-s', [], []),
        ('flag', '--s', [], []),
        # A used flag should not show in completions
        (
            'flag -n',
            '--',
            ['--append_const_flag', '--append_flag', '--count_flag', '--help', '--remainder_flag', '--required_flag'],
            [
                '--required_flag',
                '[--append_const_flag]',
                '[--append_flag]',
                '[--count_flag]',
                '[--help]',
                '[--remainder_flag]',
            ],
        ),
        # Flags with actions set to append, append_const, and count will always show even if they've been used
        (
            'flag --append_const_flag -c --append_flag value',
            '--',
            [
                '--append_const_flag',
                '--append_flag',
                '--count_flag',
                '--help',
                '--normal_flag',
                '--remainder_flag',
                '--required_flag',
            ],
            [
                '--required_flag',
                '[--append_const_flag]',
                '[--append_flag]',
                '[--count_flag]',
                '[--help]',
                '[--normal_flag]',
                '[--remainder_flag]',
            ],
        ),
        # Non-default flag prefix character (+)
        (
            'plus_flag',
            '+',
            ['++help', '++normal_flag', '+h', '+n', '+q', '++required_flag'],
            ['+q, ++required_flag', '[+h, ++help]', '[+n, ++normal_flag]'],
        ),
        (
            'plus_flag',
            '++',
            ['++help', '++normal_flag', '++required_flag'],
            ['++required_flag', '[++help]', '[++normal_flag]'],
        ),
        # Flag completion should not occur after '--' since that tells argparse all remaining arguments are non-flags
        ('flag --', '--', [], []),
        ('flag --help --', '--', [], []),
        ('plus_flag --', '++', [], []),
        ('plus_flag ++help --', '++', [], []),
        # Test remaining flag names complete after all positionals are complete
        ('pos_and_flag', '', ['a', 'choice'], ['a', 'choice']),
        ('pos_and_flag choice ', '', ['--flag', '--help', '-f', '-h'], ['[-f, --flag]', '[-h, --help]']),
        ('pos_and_flag choice -f ', '', ['--help', '-h'], ['[-h, --help]']),
        ('pos_and_flag choice -f -h ', '', [], []),
    ],
)
def test_autcomp_flag_completion(ac_app, command_and_args, text, completion_matches, display_matches) -> None:
    line = f'{command_and_args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completion_matches:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completion_matches, key=ac_app.default_sort_key)
    assert ac_app.display_matches == sorted(display_matches, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('flag', 'text', 'completions'),
    [
        ('-l', '', ArgparseCompleterTester.static_choices_list),
        ('--list', 's', ['static', 'stop']),
        ('-p', '', ArgparseCompleterTester.choices_from_provider),
        ('--provider', 'pr', ['provider', 'probably']),
        ('-n', '', ArgparseCompleterTester.num_choices),
        ('--num', '1', ['1 ']),
        ('--num', '-', [-1, -2, -12]),
        ('--num', '-1', [-1, -12]),
        ('--num_completion_items', '', ArgparseCompleterTester.num_completion_items),
    ],
)
def test_autocomp_flag_choices_completion(ac_app, flag, text, completions) -> None:
    line = f'choices {flag} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    # Numbers will be sorted in ascending order and then converted to strings by ArgparseCompleter
    if completions and all(isinstance(x, numbers.Number) for x in completions):
        completions = [str(x) for x in sorted(completions)]
    else:
        completions = sorted(completions, key=ac_app.default_sort_key)

    assert ac_app.completion_matches == completions


@pytest.mark.parametrize(
    ('pos', 'text', 'completions'),
    [
        (1, '', ArgparseCompleterTester.static_choices_list),
        (1, 's', ['static', 'stop']),
        (2, '', ArgparseCompleterTester.choices_from_provider),
        (2, 'pr', ['provider', 'probably']),
        (3, '', ArgparseCompleterTester.non_negative_num_choices),
        (3, '2', [2, 22]),
        (4, '', []),
    ],
)
def test_autocomp_positional_choices_completion(ac_app, pos, text, completions) -> None:
    # Generate line were preceding positionals are already filled
    line = 'choices {} {}'.format('foo ' * (pos - 1), text)
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    # Numbers will be sorted in ascending order and then converted to strings by ArgparseCompleter
    if completions and all(isinstance(x, numbers.Number) for x in completions):
        completions = [str(x) for x in sorted(completions)]
    else:
        completions = sorted(completions, key=ac_app.default_sort_key)

    assert ac_app.completion_matches == completions


def test_flag_sorting(ac_app) -> None:
    # This test exercises the case where a positional arg has non-negative integers for its choices.
    # ArgparseCompleter will sort these numerically before converting them to strings. As a result,
    # cmd2.matches_sorted gets set to True. If no completion matches are returned and the entered
    # text looks like the beginning of a flag (e.g -), then ArgparseCompleter will try to complete
    # flag names next. Before it does this, cmd2.matches_sorted is reset to make sure the flag names
    # get sorted correctly.
    option_strings = []
    for action in ac_app.choices_parser._actions:
        option_strings.extend(action.option_strings)
    option_strings.sort(key=ac_app.default_sort_key)

    text = '-'
    line = f'choices arg1 arg2 arg3 {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    assert first_match is not None
    assert ac_app.completion_matches == option_strings


@pytest.mark.parametrize(
    ('flag', 'text', 'completions'),
    [('-c', '', ArgparseCompleterTester.completions_for_flag), ('--completer', 'f', ['flag', 'fairly'])],
)
def test_autocomp_flag_completers(ac_app, flag, text, completions) -> None:
    line = f'completer {flag} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('pos', 'text', 'completions'),
    [
        (1, '', ArgparseCompleterTester.completions_for_pos_1),
        (1, 'p', ['positional_1', 'probably']),
        (2, '', ArgparseCompleterTester.completions_for_pos_2),
        (2, 'm', ['missed', 'me']),
    ],
)
def test_autocomp_positional_completers(ac_app, pos, text, completions) -> None:
    # Generate line were preceding positionals are already filled
    line = 'completer {} {}'.format('foo ' * (pos - 1), text)
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


def test_autocomp_blank_token(ac_app) -> None:
    """Force a blank token to make sure ArgparseCompleter consumes them like argparse does"""
    from cmd2.argparse_completer import (
        ArgparseCompleter,
    )

    blank = ''

    # Blank flag arg will be consumed. Therefore we expect to be completing the first positional.
    text = ''
    line = f'completer -c {blank} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    completer = ArgparseCompleter(ac_app.completer_parser, ac_app)
    tokens = ['-c', blank, text]
    completions = completer.complete(text, line, begidx, endidx, tokens)
    assert sorted(completions) == sorted(ArgparseCompleterTester.completions_for_pos_1)

    # Blank arg for first positional will be consumed. Therefore we expect to be completing the second positional.
    text = ''
    line = f'completer {blank} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    completer = ArgparseCompleter(ac_app.completer_parser, ac_app)
    tokens = [blank, text]
    completions = completer.complete(text, line, begidx, endidx, tokens)
    assert sorted(completions) == sorted(ArgparseCompleterTester.completions_for_pos_2)


@with_ansi_style(ru.AllowStyle.ALWAYS)
def test_completion_items(ac_app) -> None:
    # First test CompletionItems created from strings
    text = ''
    line = f'choices --completion_items {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    assert first_match is not None
    assert len(ac_app.completion_matches) == len(ac_app.completion_item_choices)
    assert len(ac_app.display_matches) == len(ac_app.completion_item_choices)

    lines = ac_app.formatted_completions.splitlines()

    # Since the CompletionItems were created from strings, the left-most column is left-aligned.
    # Therefore choice_1 will begin the line (with 1 space for padding).
    assert lines[2].startswith(' choice_1')
    assert lines[2].strip().endswith('Description 1')

    # Verify that the styled string was converted to a Rich Text object so that
    # Rich could correctly calculate its display width. Since it was the longest
    # description in the table, we should only see one space of padding after it.
    assert lines[3].endswith("\x1b[34mString with style\x1b[0m ")

    # Verify that the styled Rich Text also rendered.
    assert lines[4].endswith("\x1b[31mText with style  \x1b[0m ")

    # Now test CompletionItems created from numbers
    text = ''
    line = f'choices --num_completion_items {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    assert first_match is not None
    assert len(ac_app.completion_matches) == len(ac_app.num_completion_items)
    assert len(ac_app.display_matches) == len(ac_app.num_completion_items)

    lines = ac_app.formatted_completions.splitlines()

    # Since the CompletionItems were created from numbers, the left-most column is right-aligned.
    # Therefore 1.5 will be right-aligned.
    assert lines[2].startswith("                  1.5")
    assert lines[2].strip().endswith('One.Five')


@pytest.mark.parametrize(
    ('num_aliases', 'show_description'),
    [
        # The number of completion results determines if the description field of CompletionItems gets displayed
        # in the tab completions. The count must be greater than 1 and less than ac_app.max_completion_items,
        # which defaults to 50.
        (1, False),
        (5, True),
        (100, False),
    ],
)
def test_max_completion_items(ac_app, num_aliases, show_description) -> None:
    # Create aliases
    for i in range(num_aliases):
        run_cmd(ac_app, f'alias create fake_alias{i} help')

    assert len(ac_app.aliases) == num_aliases

    text = 'fake_alias'
    line = f'alias list {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    assert first_match is not None
    assert len(ac_app.completion_matches) == num_aliases
    assert len(ac_app.display_matches) == num_aliases

    assert bool(ac_app.formatted_completions) == show_description
    if show_description:
        # If show_description is True, the table will show both the alias name and value
        description_displayed = False
        for line in ac_app.formatted_completions.splitlines():
            if 'fake_alias0' in line and 'help' in line:
                description_displayed = True
                break

        assert description_displayed


@pytest.mark.parametrize(
    ('args', 'completions'),
    [
        # Flag with nargs = 2
        ('--set_value', ArgparseCompleterTester.set_value_choices),
        ('--set_value set', ['value', 'choices']),
        # Both args are filled. At positional arg now.
        ('--set_value set value', ArgparseCompleterTester.positional_choices),
        # Using the flag again will reset the choices available
        ('--set_value set value --set_value', ArgparseCompleterTester.set_value_choices),
        # Flag with nargs = ONE_OR_MORE
        ('--one_or_more', ArgparseCompleterTester.one_or_more_choices),
        ('--one_or_more one', ['or', 'more', 'choices']),
        # Flag with nargs = OPTIONAL
        ('--optional', ArgparseCompleterTester.optional_choices),
        # Only one arg allowed for an OPTIONAL. At positional now.
        ('--optional optional', ArgparseCompleterTester.positional_choices),
        # Flag with nargs range (1, 2)
        ('--range', ArgparseCompleterTester.range_choices),
        ('--range some', ['range', 'choices']),
        # Already used 2 args so at positional
        ('--range some range', ArgparseCompleterTester.positional_choices),
        # Flag with nargs = REMAINDER
        ('--remainder', ArgparseCompleterTester.remainder_choices),
        ('--remainder remainder ', ['choices ']),
        # No more flags can appear after a REMAINDER flag)
        ('--remainder choices --set_value', ['remainder ']),
        # Double dash ends the current flag
        ('--range choice --', ArgparseCompleterTester.positional_choices),
        # Double dash ends a REMAINDER flag
        ('--remainder remainder --', ArgparseCompleterTester.positional_choices),
        # No more flags after a double dash
        ('-- --one_or_more ', ArgparseCompleterTester.positional_choices),
        # Consume positional
        ('', ArgparseCompleterTester.positional_choices),
        ('positional', ['the', 'choices']),
        # Intermixed flag and positional
        ('positional --set_value', ArgparseCompleterTester.set_value_choices),
        ('positional --set_value set', ['choices', 'value']),
        # Intermixed flag and positional with flag finishing
        ('positional --set_value set value', ['the', 'choices']),
        ('positional --range choice --', ['the', 'choices']),
        # REMAINDER positional
        ('the positional', ArgparseCompleterTester.remainder_choices),
        ('the positional remainder', ['choices ']),
        ('the positional remainder choices', []),
        # REMAINDER positional. Flags don't work in REMAINDER
        ('the positional --set_value', ArgparseCompleterTester.remainder_choices),
        ('the positional remainder --set_value', ['choices ']),
    ],
)
def test_autcomp_nargs(ac_app, args, completions) -> None:
    text = ''
    line = f'nargs {args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('command_and_args', 'text', 'is_error'),
    [
        # Flag is finished before moving on
        ('hint --flag foo --', '', False),
        ('hint --flag foo --help', '', False),
        ('hint --flag foo', '--', False),
        ('nargs --one_or_more one --', '', False),
        ('nargs --one_or_more one or --set_value', '', False),
        ('nargs --one_or_more one or more', '--', False),
        ('nargs --set_value set value --', '', False),
        ('nargs --set_value set value --one_or_more', '', False),
        ('nargs --set_value set value', '--', False),
        ('nargs --set_val set value', '--', False),  # This exercises our abbreviated flag detection
        ('nargs --range choices --', '', False),
        ('nargs --range choices range --set_value', '', False),
        ('nargs --range range', '--', False),
        # Flag is not finished before moving on
        ('hint --flag --', '', True),
        ('hint --flag --help', '', True),
        ('hint --flag', '--', True),
        ('nargs --one_or_more --', '', True),
        ('nargs --one_or_more --set_value', '', True),
        ('nargs --one_or_more', '--', True),
        ('nargs --set_value set --', '', True),
        ('nargs --set_value set --one_or_more', '', True),
        ('nargs --set_value set', '--', True),
        ('nargs --set_val set', '--', True),  # This exercises our abbreviated flag detection
        ('nargs --range --', '', True),
        ('nargs --range --set_value', '', True),
        ('nargs --range', '--', True),
    ],
)
def test_unfinished_flag_error(ac_app, command_and_args, text, is_error, capsys) -> None:
    line = f'{command_and_args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)

    out, _err = capsys.readouterr()
    assert is_error == all(x in out for x in ["Error: argument", "expected"])


def test_completion_items_arg_header(ac_app) -> None:
    # Test when metavar is None
    text = ''
    line = f'choices --desc_header {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)
    assert "DESC_HEADER" in normalize(ac_app.formatted_completions)[0]

    # Test when metavar is a string
    text = ''
    line = f'choices --no_header {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)
    assert ac_app.STR_METAVAR in normalize(ac_app.formatted_completions)[0]

    # Test when metavar is a tuple
    text = ''
    line = f'choices --tuple_metavar {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # We are completing the first argument of this flag. The first element in the tuple should be the column header.
    complete_tester(text, line, begidx, endidx, ac_app)
    assert ac_app.TUPLE_METAVAR[0].upper() in normalize(ac_app.formatted_completions)[0]

    text = ''
    line = f'choices --tuple_metavar token_1 {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # We are completing the second argument of this flag. The second element in the tuple should be the column header.
    complete_tester(text, line, begidx, endidx, ac_app)
    assert ac_app.TUPLE_METAVAR[1].upper() in normalize(ac_app.formatted_completions)[0]

    text = ''
    line = f'choices --tuple_metavar token_1 token_2 {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # We are completing the third argument of this flag. It should still be the second tuple element
    # in the column header since the tuple only has two strings in it.
    complete_tester(text, line, begidx, endidx, ac_app)
    assert ac_app.TUPLE_METAVAR[1].upper() in normalize(ac_app.formatted_completions)[0]


def test_completion_items_descriptive_headers(ac_app) -> None:
    from cmd2.argparse_completer import (
        DEFAULT_DESCRIPTIVE_HEADERS,
    )

    # This argument provided a descriptive header
    text = ''
    line = f'choices --desc_header {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)
    assert ac_app.CUSTOM_DESC_HEADERS[0] in normalize(ac_app.formatted_completions)[0]

    # This argument did not provide a descriptive header, so it should be DEFAULT_DESCRIPTIVE_HEADERS
    text = ''
    line = f'choices --no_header {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)
    assert DEFAULT_DESCRIPTIVE_HEADERS[0] in normalize(ac_app.formatted_completions)[0]


@pytest.mark.parametrize(
    ('command_and_args', 'text', 'has_hint'),
    [
        # Normal cases
        ('hint', '', True),
        ('hint --flag', '', True),
        ('hint --suppressed_help', '', False),
        ('hint --suppressed_hint', '', False),
        # Hint because flag does not have enough values to be considered finished
        ('nargs --one_or_more', '-', True),
        # This flag has reached its minimum value count and therefore a new flag could start.
        # However the flag can still consume values and the text is not a single prefix character.
        # Therefore a hint will be shown.
        ('nargs --one_or_more choices', 'bad_completion', True),
        # Like the previous case, but this time text is a single prefix character which will cause flag
        # name completion to occur instead of a hint for the current flag.
        ('nargs --one_or_more choices', '-', False),
        # Hint because this is a REMAINDER flag and therefore no more flag name completions occur.
        ('nargs --remainder', '-', True),
        # No hint for the positional because text is a single prefix character which results in flag name completion
        ('hint', '-', False),
        # Hint because this is a REMAINDER positional and therefore no more flag name completions occur.
        ('nargs the choices', '-', True),
        ('nargs the choices remainder', '-', True),
    ],
)
def test_autocomp_hint(ac_app, command_and_args, text, has_hint, capsys) -> None:
    line = f'{command_and_args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    complete_tester(text, line, begidx, endidx, ac_app)
    out, _err = capsys.readouterr()
    if has_hint:
        assert "Hint:\n" in out
    else:
        assert not out


def test_autocomp_hint_no_help_text(ac_app, capsys) -> None:
    text = ''
    line = f'hint foo {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    out, _err = capsys.readouterr()

    assert first_match is None
    assert out != '''\nHint:\n  NO_HELP_POS\n\n'''


@pytest.mark.parametrize(
    ('args', 'text'),
    [
        # Exercise a flag arg and choices function that raises a CompletionError
        ('--choice ', 'choice'),
        # Exercise a positional arg and completer that raises a CompletionError
        ('', 'completer'),
    ],
)
def test_completion_error(ac_app, capsys, args, text) -> None:
    line = f'raise_completion_error {args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    out, _err = capsys.readouterr()

    assert first_match is None
    assert f"{text} broke something" in out


@pytest.mark.parametrize(
    ('command_and_args', 'completions'),
    [
        # Exercise a choices function that receives arg_tokens dictionary
        ('arg_tokens choice subcmd', ['choice', 'subcmd']),
        # Exercise a completer that receives arg_tokens dictionary
        ('arg_tokens completer subcmd fake', ['completer', 'subcmd']),
        # Exercise overriding parent_arg from the subcommand
        ('arg_tokens completer subcmd --parent_arg override fake', ['override', 'subcmd']),
    ],
)
def test_arg_tokens(ac_app, command_and_args, completions) -> None:
    text = ''
    line = f'{command_and_args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    if completions:
        assert first_match is not None
    else:
        assert first_match is None

    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


@pytest.mark.parametrize(
    ('command_and_args', 'text', 'output_contains', 'first_match'),
    [
        # Group isn't done. Hint will show for optional positional and no completions returned
        ('mutex', '', 'the optional positional', None),
        # Group isn't done. Flag name will still complete.
        ('mutex', '--fl', '', '--flag '),
        # Group isn't done. Flag hint will show.
        ('mutex --flag', '', 'the flag arg', None),
        # Group finished by optional positional. No flag name will complete.
        ('mutex pos_val', '--fl', '', None),
        # Group finished by optional positional. Error will display trying to complete the flag's value.
        ('mutex pos_val --flag', '', 'f/--flag: not allowed with argument optional_pos', None),
        # Group finished by --flag. Optional positional will be skipped and last_arg will show its hint.
        ('mutex --flag flag_val', '', 'the last arg', None),
        # Group finished by --flag. Other flag name won't complete.
        ('mutex --flag flag_val', '--oth', '', None),
        # Group finished by --flag. Error will display trying to complete other flag's value.
        ('mutex --flag flag_val --other', '', '-o/--other_flag: not allowed with argument -f/--flag', None),
        # Group finished by --flag. That same flag can be used again so it's hint will show.
        ('mutex --flag flag_val --flag', '', 'the flag arg', None),
    ],
)
def test_complete_mutex_group(ac_app, command_and_args, text, output_contains, first_match, capsys) -> None:
    line = f'{command_and_args} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert first_match == complete_tester(text, line, begidx, endidx, ac_app)

    out, _err = capsys.readouterr()
    assert output_contains in out


def test_single_prefix_char() -> None:
    from cmd2.argparse_completer import (
        _single_prefix_char,
    )

    parser = Cmd2ArgumentParser(prefix_chars='-+')

    # Invalid
    assert not _single_prefix_char('', parser)
    assert not _single_prefix_char('--', parser)
    assert not _single_prefix_char('-+', parser)
    assert not _single_prefix_char('++has space', parser)
    assert not _single_prefix_char('foo', parser)

    # Valid
    assert _single_prefix_char('-', parser)
    assert _single_prefix_char('+', parser)


def test_looks_like_flag() -> None:
    from cmd2.argparse_completer import (
        _looks_like_flag,
    )

    parser = Cmd2ArgumentParser()

    # Does not start like a flag
    assert not _looks_like_flag('', parser)
    assert not _looks_like_flag('non-flag', parser)
    assert not _looks_like_flag('-', parser)
    assert not _looks_like_flag('--has space', parser)
    assert not _looks_like_flag('-2', parser)

    # Does start like a flag
    assert _looks_like_flag('--', parser)
    assert _looks_like_flag('-flag', parser)
    assert _looks_like_flag('--flag', parser)


def test_complete_command_no_tokens(ac_app) -> None:
    from cmd2.argparse_completer import (
        ArgparseCompleter,
    )

    parser = Cmd2ArgumentParser()
    ac = ArgparseCompleter(parser, ac_app)

    completions = ac.complete(text='', line='', begidx=0, endidx=0, tokens=[])
    assert not completions


def test_complete_command_help_no_tokens(ac_app) -> None:
    from cmd2.argparse_completer import (
        ArgparseCompleter,
    )

    parser = Cmd2ArgumentParser()
    ac = ArgparseCompleter(parser, ac_app)

    completions = ac.complete_subcommand_help(text='', line='', begidx=0, endidx=0, tokens=[])
    assert not completions


@pytest.mark.parametrize(
    ('flag', 'completions'), [('--provider', standalone_choices), ('--completer', standalone_completions)]
)
def test_complete_standalone(ac_app, flag, completions) -> None:
    text = ''
    line = f'standalone {flag} {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, ac_app)
    assert first_match is not None
    assert ac_app.completion_matches == sorted(completions, key=ac_app.default_sort_key)


# Custom ArgparseCompleter-based class
class CustomCompleter(argparse_completer.ArgparseCompleter):
    def _complete_flags(self, text: str, line: str, begidx: int, endidx: int, matched_flags: list[str]) -> list[str]:
        """Override so flags with 'complete_when_ready' set to True will complete only when app is ready"""
        # Find flags which should not be completed and place them in matched_flags
        for flag in self._flags:
            action = self._flag_to_action[flag]
            app: CustomCompleterApp = cast(CustomCompleterApp, self._cmd2_app)
            if action.get_complete_when_ready() is True and not app.is_ready:
                matched_flags.append(flag)

        return super()._complete_flags(text, line, begidx, endidx, matched_flags)


# Add a custom argparse action attribute
argparse_custom.register_argparse_argument_parameter('complete_when_ready', bool)


# App used to test custom ArgparseCompleter types and custom argparse attributes
class CustomCompleterApp(cmd2.Cmd):
    def __init__(self) -> None:
        super().__init__()
        self.is_ready = True

    # Parser that's used to test setting the app-wide default ArgparseCompleter type
    default_completer_parser = Cmd2ArgumentParser(description="Testing app-wide argparse completer")
    default_completer_parser.add_argument('--myflag', complete_when_ready=True)

    @with_argparser(default_completer_parser)
    def do_default_completer(self, args: argparse.Namespace) -> None:
        """Test command"""

    # Parser that's used to test setting a custom completer at the parser level
    custom_completer_parser = Cmd2ArgumentParser(
        description="Testing parser-specific argparse completer", ap_completer_type=CustomCompleter
    )
    custom_completer_parser.add_argument('--myflag', complete_when_ready=True)

    @with_argparser(custom_completer_parser)
    def do_custom_completer(self, args: argparse.Namespace) -> None:
        """Test command"""

    # Test as_subcommand_to decorator with custom completer
    top_parser = Cmd2ArgumentParser(description="Top Command")
    top_parser.add_subparsers(dest='subcommand', metavar='SUBCOMMAND', required=True)

    @with_argparser(top_parser)
    def do_top(self, args: argparse.Namespace) -> None:
        """Top level command"""
        # Call handler for whatever subcommand was selected
        handler = args.cmd2_handler.get()
        handler(args)

    # Parser for a subcommand with no custom completer type
    no_custom_completer_parser = Cmd2ArgumentParser(description="No custom completer")
    no_custom_completer_parser.add_argument('--myflag', complete_when_ready=True)

    @cmd2.as_subcommand_to('top', 'no_custom', no_custom_completer_parser, help="no custom completer")
    def _subcmd_no_custom(self, args: argparse.Namespace) -> None:
        pass

    # Parser for a subcommand with a custom completer type
    custom_completer_parser2 = Cmd2ArgumentParser(description="Custom completer", ap_completer_type=CustomCompleter)
    custom_completer_parser2.add_argument('--myflag', complete_when_ready=True)

    @cmd2.as_subcommand_to('top', 'custom', custom_completer_parser2, help="custom completer")
    def _subcmd_custom(self, args: argparse.Namespace) -> None:
        pass


@pytest.fixture
def custom_completer_app():
    return CustomCompleterApp()


def test_default_custom_completer_type(custom_completer_app: CustomCompleterApp) -> None:
    """Test altering the app-wide default ArgparseCompleter type"""
    try:
        argparse_completer.set_default_ap_completer_type(CustomCompleter)

        text = '--m'
        line = f'default_completer {text}'
        endidx = len(line)
        begidx = endidx - len(text)

        # The flag should complete because app is ready
        custom_completer_app.is_ready = True
        assert complete_tester(text, line, begidx, endidx, custom_completer_app) is not None
        assert custom_completer_app.completion_matches == ['--myflag ']

        # The flag should not complete because app is not ready
        custom_completer_app.is_ready = False
        assert complete_tester(text, line, begidx, endidx, custom_completer_app) is None
        assert not custom_completer_app.completion_matches

    finally:
        # Restore the default completer
        argparse_completer.set_default_ap_completer_type(argparse_completer.ArgparseCompleter)


def test_custom_completer_type(custom_completer_app: CustomCompleterApp) -> None:
    """Test parser with a specific custom ArgparseCompleter type"""
    text = '--m'
    line = f'custom_completer {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # The flag should complete because app is ready
    custom_completer_app.is_ready = True
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is not None
    assert custom_completer_app.completion_matches == ['--myflag ']

    # The flag should not complete because app is not ready
    custom_completer_app.is_ready = False
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is None
    assert not custom_completer_app.completion_matches


def test_decorated_subcmd_custom_completer(custom_completer_app: CustomCompleterApp) -> None:
    """Tests custom completer type on a subcommand created with @cmd2.as_subcommand_to"""
    # First test the subcommand without the custom completer
    text = '--m'
    line = f'top no_custom {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # The flag should complete regardless of ready state since this subcommand isn't using the custom completer
    custom_completer_app.is_ready = True
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is not None
    assert custom_completer_app.completion_matches == ['--myflag ']

    custom_completer_app.is_ready = False
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is not None
    assert custom_completer_app.completion_matches == ['--myflag ']

    # Now test the subcommand with the custom completer
    text = '--m'
    line = f'top custom {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # The flag should complete because app is ready
    custom_completer_app.is_ready = True
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is not None
    assert custom_completer_app.completion_matches == ['--myflag ']

    # The flag should not complete because app is not ready
    custom_completer_app.is_ready = False
    assert complete_tester(text, line, begidx, endidx, custom_completer_app) is None
    assert not custom_completer_app.completion_matches


def test_add_parser_custom_completer() -> None:
    """Tests setting a custom completer type on a subcommand using add_parser()"""
    parser = Cmd2ArgumentParser()
    subparsers = parser.add_subparsers()

    no_custom_completer_parser = subparsers.add_parser(name="no_custom_completer")
    assert no_custom_completer_parser.get_ap_completer_type() is None  # type: ignore[attr-defined]

    custom_completer_parser = subparsers.add_parser(name="custom_completer", ap_completer_type=CustomCompleter)
    assert custom_completer_parser.get_ap_completer_type() is CustomCompleter  # type: ignore[attr-defined]



"""Simple example demonstrating basic CommandSet usage."""

from typing import Any

import cmd2
from cmd2 import (
    CommandSet,
    with_default_category,
)


@with_default_category('Default Category')
class MyBaseCommandSet(CommandSet):
    """Defines a default category for all sub-class CommandSets"""

    def __init__(self, _: Any) -> None:
        super().__init__()


class ChildInheritsParentCategories(MyBaseCommandSet):
    """This subclass doesn't declare any categories so all commands here are also categorized under 'Default Category'"""

    def do_hello(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Hello')

    def do_world(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('World')


@with_default_category('Non-Heritable Category', heritable=False)
class ChildOverridesParentCategoriesNonHeritable(MyBaseCommandSet):
    """This subclass overrides the 'Default Category' from the parent, but in a non-heritable fashion. Sub-classes of this
    CommandSet will not inherit this category and will, instead, inherit 'Default Category'
    """

    def do_goodbye(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Goodbye')


class GrandchildInheritsGrandparentCategory(ChildOverridesParentCategoriesNonHeritable):
    """This subclass's parent class declared its default category non-heritable. Instead, it inherits the category defined
    by the grandparent class.
    """

    def do_aloha(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Aloha')


@with_default_category('Heritable Category')
class ChildOverridesParentCategories(MyBaseCommandSet):
    """This subclass is decorated with a default category that is heritable. This overrides the parent class's default
    category declaration.
    """

    def do_bonjour(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Bonjour')


class GrandchildInheritsHeritable(ChildOverridesParentCategories):
    """This subclass's parent declares a default category that overrides its parent. As a result, commands in this
    CommandSet will be categorized under 'Heritable Category'
    """

    def do_monde(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Monde')


class ExampleApp(cmd2.Cmd):
    """Example to demonstrate heritable default categories"""

    def __init__(self) -> None:
        super().__init__(auto_load_commands=False)

    def do_something(self, arg) -> None:
        self.poutput('this is the something command')


def test_heritable_categories() -> None:
    app = ExampleApp()

    base_cs = MyBaseCommandSet(0)
    assert getattr(base_cs, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) == 'Default Category'

    child1 = ChildInheritsParentCategories(1)
    assert getattr(child1, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) == 'Default Category'
    app.register_command_set(child1)
    assert getattr(app.cmd_func('hello').__func__, cmd2.constants.CMD_ATTR_HELP_CATEGORY, None) == 'Default Category'
    app.unregister_command_set(child1)

    child_nonheritable = ChildOverridesParentCategoriesNonHeritable(2)
    assert getattr(child_nonheritable, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) != 'Non-Heritable Category'
    app.register_command_set(child_nonheritable)
    assert getattr(app.cmd_func('goodbye').__func__, cmd2.constants.CMD_ATTR_HELP_CATEGORY, None) == 'Non-Heritable Category'
    app.unregister_command_set(child_nonheritable)

    grandchild1 = GrandchildInheritsGrandparentCategory(3)
    assert getattr(grandchild1, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) == 'Default Category'
    app.register_command_set(grandchild1)
    assert getattr(app.cmd_func('aloha').__func__, cmd2.constants.CMD_ATTR_HELP_CATEGORY, None) == 'Default Category'
    app.unregister_command_set(grandchild1)

    child_overrides = ChildOverridesParentCategories(4)
    assert getattr(child_overrides, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) == 'Heritable Category'
    app.register_command_set(child_overrides)
    assert getattr(app.cmd_func('bonjour').__func__, cmd2.constants.CMD_ATTR_HELP_CATEGORY, None) == 'Heritable Category'
    app.unregister_command_set(child_overrides)

    grandchild2 = GrandchildInheritsHeritable(5)
    assert getattr(grandchild2, cmd2.constants.CLASS_ATTR_DEFAULT_HELP_CATEGORY, None) == 'Heritable Category'



"""Test CommandSet"""

import argparse
import signal

import pytest

import cmd2
from cmd2 import (
    Settable,
)
from cmd2.exceptions import (
    CommandSetRegistrationError,
)

from .conftest import (
    WithCommandSets,
    complete_tester,
    normalize,
    run_cmd,
)


class CommandSetBase(cmd2.CommandSet):
    pass


@cmd2.with_default_category('Fruits')
class CommandSetA(CommandSetBase):
    def on_register(self, cmd) -> None:
        super().on_register(cmd)
        print("in on_register now")

    def on_registered(self) -> None:
        super().on_registered()
        print("in on_registered now")

    def on_unregister(self) -> None:
        super().on_unregister()
        print("in on_unregister now")

    def on_unregistered(self) -> None:
        super().on_unregistered()
        print("in on_unregistered now")

    def do_apple(self, statement: cmd2.Statement) -> None:
        self._cmd.poutput('Apple!')

    def do_banana(self, statement: cmd2.Statement) -> None:
        """Banana Command"""
        self._cmd.poutput('Banana!!')

    cranberry_parser = cmd2.Cmd2ArgumentParser()
    cranberry_parser.add_argument('arg1', choices=['lemonade', 'juice', 'sauce'])

    @cmd2.with_argparser(cranberry_parser, with_unknown_args=True)
    def do_cranberry(self, ns: argparse.Namespace, unknown: list[str]) -> None:
        self._cmd.poutput(f'Cranberry {ns.arg1}!!')
        if unknown and len(unknown):
            self._cmd.poutput('Unknown: ' + ', '.join(['{}'] * len(unknown)).format(*unknown))
        self._cmd.last_result = {'arg1': ns.arg1, 'unknown': unknown}

    def help_cranberry(self) -> None:
        self._cmd.stdout.write('This command does diddly squat...\n')

    @cmd2.with_argument_list
    @cmd2.with_category('Also Alone')
    def do_durian(self, args: list[str]) -> None:
        """Durian Command"""
        self._cmd.poutput(f'{len(args)} Arguments: ')
        self._cmd.poutput(', '.join(['{}'] * len(args)).format(*args))
        self._cmd.last_result = {'args': args}

    def complete_durian(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self._cmd.basic_complete(text, line, begidx, endidx, ['stinks', 'smells', 'disgusting'])

    elderberry_parser = cmd2.Cmd2ArgumentParser()
    elderberry_parser.add_argument('arg1')

    @cmd2.with_category('Alone')
    @cmd2.with_argparser(elderberry_parser)
    def do_elderberry(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput(f'Elderberry {ns.arg1}!!')
        self._cmd.last_result = {'arg1': ns.arg1}

    # Test that CommandSet with as_subcommand_to decorator successfully loads
    # during `cmd2.Cmd.__init__()`.
    main_parser = cmd2.Cmd2ArgumentParser(description="Main Command")
    main_parser.add_subparsers(dest='subcommand', metavar='SUBCOMMAND', required=True)

    @cmd2.with_category('Alone')
    @cmd2.with_argparser(main_parser)
    def do_main(self, args: argparse.Namespace) -> None:
        # Call handler for whatever subcommand was selected
        handler = args.cmd2_handler.get()
        handler(args)

    # main -> sub
    subcmd_parser = cmd2.Cmd2ArgumentParser(description="Sub Command")

    @cmd2.as_subcommand_to('main', 'sub', subcmd_parser, help="sub command")
    def subcmd_func(self, args: argparse.Namespace) -> None:
        self._cmd.poutput("Subcommand Ran")


@cmd2.with_default_category('Command Set B')
class CommandSetB(CommandSetBase):
    def __init__(self, arg1) -> None:
        super().__init__()
        self._arg1 = arg1

    def do_aardvark(self, statement: cmd2.Statement) -> None:
        self._cmd.poutput('Aardvark!')

    def do_bat(self, statement: cmd2.Statement) -> None:
        """Banana Command"""
        self._cmd.poutput('Bat!!')

    def do_crocodile(self, statement: cmd2.Statement) -> None:
        self._cmd.poutput('Crocodile!!')


def test_autoload_commands(autoload_command_sets_app) -> None:
    # verifies that, when autoload is enabled, CommandSets and registered functions all show up

    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = autoload_command_sets_app._build_command_info()

    assert 'Alone' in cmds_cats
    assert 'elderberry' in cmds_cats['Alone']
    assert 'main' in cmds_cats['Alone']

    # Test subcommand was autoloaded
    result = autoload_command_sets_app.app_cmd('main sub')
    assert 'Subcommand Ran' in result.stdout

    assert 'Also Alone' in cmds_cats
    assert 'durian' in cmds_cats['Also Alone']

    assert 'Fruits' in cmds_cats
    assert 'cranberry' in cmds_cats['Fruits']

    assert 'Command Set B' not in cmds_cats


def test_command_synonyms() -> None:
    """Test the use of command synonyms in CommandSets"""

    class SynonymCommandSet(cmd2.CommandSet):
        def __init__(self, arg1) -> None:
            super().__init__()
            self._arg1 = arg1

        @cmd2.with_argparser(cmd2.Cmd2ArgumentParser(description="Native Command"))
        def do_builtin(self, _) -> None:
            pass

        # Create a synonym to a command inside of this CommandSet
        do_builtin_synonym = do_builtin

        # Create a synonym to a command outside of this CommandSet with subcommands.
        # This will best test the synonym check in cmd2.Cmd._check_uninstallable() when
        # we unresgister this CommandSet.
        do_alias_synonym = cmd2.Cmd.do_alias

    cs = SynonymCommandSet("foo")
    app = WithCommandSets(command_sets=[cs])

    # Make sure the synonyms have the same parser as what they alias
    builtin_parser = app._command_parsers.get(app.do_builtin)
    builtin_synonym_parser = app._command_parsers.get(app.do_builtin_synonym)
    assert builtin_parser is not None
    assert builtin_parser is builtin_synonym_parser

    alias_parser = app._command_parsers.get(cmd2.Cmd.do_alias)
    alias_synonym_parser = app._command_parsers.get(app.do_alias_synonym)
    assert alias_parser is not None
    assert alias_parser is alias_synonym_parser

    # Unregister the CommandSet and make sure built-in command and synonyms are gone
    app.unregister_command_set(cs)
    assert not hasattr(app, "do_builtin")
    assert not hasattr(app, "do_builtin_synonym")
    assert not hasattr(app, "do_alias_synonym")

    # Make sure the alias command still exists, has the same parser, and works.
    assert alias_parser is app._command_parsers.get(cmd2.Cmd.do_alias)
    out, _err = run_cmd(app, 'alias --help')
    assert normalize(alias_parser.format_help())[0] in out


def test_custom_construct_commandsets() -> None:
    command_set_b = CommandSetB('foo')

    # Verify that _cmd cannot be accessed until CommandSet is registered.
    with pytest.raises(CommandSetRegistrationError) as excinfo:
        command_set_b._cmd.poutput("test")
    assert "is not registered" in str(excinfo.value)

    # Verifies that a custom initialized CommandSet loads correctly when passed into the constructor
    app = WithCommandSets(command_sets=[command_set_b])

    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = app._build_command_info()
    assert 'Command Set B' in cmds_cats

    # Verifies that the same CommandSet cannot be loaded twice
    command_set_2 = CommandSetB('bar')
    with pytest.raises(CommandSetRegistrationError):
        assert app.register_command_set(command_set_2)

    # Verify that autoload doesn't conflict with a manually loaded CommandSet that could be autoloaded.
    command_set_a = CommandSetA()
    app2 = WithCommandSets(command_sets=[command_set_a])

    with pytest.raises(CommandSetRegistrationError):
        app2.register_command_set(command_set_b)

    app.unregister_command_set(command_set_b)

    app2.register_command_set(command_set_b)

    assert hasattr(app2, 'do_apple')
    assert hasattr(app2, 'do_aardvark')

    assert app2.find_commandset_for_command('aardvark') is command_set_b
    assert app2.find_commandset_for_command('apple') is command_set_a

    matches = app2.find_commandsets(CommandSetBase, subclass_match=True)
    assert command_set_a in matches
    assert command_set_b in matches
    assert command_set_2 not in matches


def test_load_commands(manual_command_sets_app, capsys) -> None:
    # now install a command set and verify the commands are now present
    cmd_set = CommandSetA()

    assert manual_command_sets_app.find_commandset_for_command('elderberry') is None
    assert not manual_command_sets_app.find_commandsets(CommandSetA)

    manual_command_sets_app.register_command_set(cmd_set)

    assert manual_command_sets_app.find_commandsets(CommandSetA)[0] is cmd_set
    assert manual_command_sets_app.find_commandset_for_command('elderberry') is cmd_set

    out = manual_command_sets_app.app_cmd('apple')
    assert 'Apple!' in out.stdout

    # Make sure registration callbacks ran
    out, _err = capsys.readouterr()
    assert "in on_register now" in out
    assert "in on_registered now" in out

    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()

    assert 'Alone' in cmds_cats
    assert 'elderberry' in cmds_cats['Alone']
    assert 'main' in cmds_cats['Alone']

    # Test subcommand was loaded
    result = manual_command_sets_app.app_cmd('main sub')
    assert 'Subcommand Ran' in result.stdout

    assert 'Fruits' in cmds_cats
    assert 'cranberry' in cmds_cats['Fruits']

    # uninstall the command set and verify it is now also no longer accessible
    manual_command_sets_app.unregister_command_set(cmd_set)

    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()

    assert 'Alone' not in cmds_cats
    assert 'Fruits' not in cmds_cats

    # Make sure unregistration callbacks ran
    out, _err = capsys.readouterr()
    assert "in on_unregister now" in out
    assert "in on_unregistered now" in out

    # uninstall a second time and verify no errors happen
    manual_command_sets_app.unregister_command_set(cmd_set)

    # reinstall the command set and verify it is accessible
    manual_command_sets_app.register_command_set(cmd_set)

    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()

    assert 'Alone' in cmds_cats
    assert 'elderberry' in cmds_cats['Alone']
    assert 'main' in cmds_cats['Alone']

    # Test subcommand was loaded
    result = manual_command_sets_app.app_cmd('main sub')
    assert 'Subcommand Ran' in result.stdout

    assert 'Fruits' in cmds_cats
    assert 'cranberry' in cmds_cats['Fruits']


def test_commandset_decorators(autoload_command_sets_app) -> None:
    result = autoload_command_sets_app.app_cmd('cranberry juice extra1 extra2')
    assert result is not None
    assert result.data is not None
    assert len(result.data['unknown']) == 2
    assert 'extra1' in result.data['unknown']
    assert 'extra2' in result.data['unknown']
    assert result.data['arg1'] == 'juice'
    assert not result.stderr

    result = autoload_command_sets_app.app_cmd('durian juice extra1 extra2')
    assert len(result.data['args']) == 3
    assert 'juice' in result.data['args']
    assert 'extra1' in result.data['args']
    assert 'extra2' in result.data['args']
    assert not result.stderr

    result = autoload_command_sets_app.app_cmd('durian')
    assert len(result.data['args']) == 0
    assert not result.stderr

    result = autoload_command_sets_app.app_cmd('elderberry')
    assert 'arguments are required' in result.stderr
    assert result.data is None

    result = autoload_command_sets_app.app_cmd('elderberry a b')
    assert 'unrecognized arguments' in result.stderr
    assert result.data is None


def test_load_commandset_errors(manual_command_sets_app, capsys) -> None:
    cmd_set = CommandSetA()

    # create a conflicting command before installing CommandSet to verify rollback behavior
    manual_command_sets_app._install_command_function('do_durian', cmd_set.do_durian)
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.register_command_set(cmd_set)

    # verify that the commands weren't installed
    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()

    assert 'Alone' not in cmds_cats
    assert 'Fruits' not in cmds_cats
    assert not manual_command_sets_app._installed_command_sets

    delattr(manual_command_sets_app, 'do_durian')

    # pre-create intentionally conflicting macro and alias names
    manual_command_sets_app.app_cmd('macro create apple run_pyscript')
    manual_command_sets_app.app_cmd('alias create banana run_pyscript')

    # now install a command set and verify the commands are now present
    manual_command_sets_app.register_command_set(cmd_set)
    _out, err = capsys.readouterr()

    # verify aliases and macros are deleted with warning if they conflict with a command
    assert "Deleting alias 'banana'" in err
    assert "Deleting macro 'apple'" in err

    # verify command functions which don't start with "do_" raise an exception
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_command_function('new_cmd', cmd_set.do_banana)

    # verify methods which don't start with "do_" raise an exception
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_command_function('do_new_cmd', cmd_set.on_register)

    # verify duplicate commands are detected
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_command_function('do_banana', cmd_set.do_banana)

    # verify bad command names are detected
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_command_function('do_bad command', cmd_set.do_banana)

    # verify error conflict with existing completer function
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_completer_function('durian', cmd_set.complete_durian)

    # verify error conflict with existing help function
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._install_help_function('cranberry', cmd_set.help_cranberry)


class LoadableBase(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        super().__init__()
        self._dummy = dummy  # prevents autoload
        self._cut_called = False

    cut_parser = cmd2.Cmd2ArgumentParser()
    cut_subparsers = cut_parser.add_subparsers(title='item', help='item to cut')

    def namespace_provider(self) -> argparse.Namespace:
        ns = argparse.Namespace()
        ns.cut_called = self._cut_called
        return ns

    @cmd2.with_argparser(cut_parser)
    def do_cut(self, ns: argparse.Namespace) -> None:
        """Cut something"""
        handler = ns.cmd2_handler.get()
        if handler is not None:
            # Call whatever subcommand function was selected
            handler(ns)
            self._cut_called = True
        else:
            # No subcommand was provided, so call help
            self._cmd.pwarning('This command does nothing without sub-parsers registered')
            self._cmd.do_help('cut')

    stir_parser = cmd2.Cmd2ArgumentParser()
    stir_subparsers = stir_parser.add_subparsers(title='item', help='what to stir')

    @cmd2.with_argparser(stir_parser, ns_provider=namespace_provider)
    def do_stir(self, ns: argparse.Namespace) -> None:
        """Stir something"""
        if not ns.cut_called:
            self._cmd.poutput('Need to cut before stirring')
            return

        handler = ns.cmd2_handler.get()
        if handler is not None:
            # Call whatever subcommand function was selected
            handler(ns)
        else:
            # No subcommand was provided, so call help
            self._cmd.pwarning('This command does nothing without sub-parsers registered')
            self._cmd.do_help('stir')

    stir_pasta_parser = cmd2.Cmd2ArgumentParser()
    stir_pasta_parser.add_argument('--option', '-o')
    stir_pasta_parser.add_subparsers(title='style', help='Stir style')

    @cmd2.as_subcommand_to('stir', 'pasta', stir_pasta_parser)
    def stir_pasta(self, ns: argparse.Namespace) -> None:
        handler = ns.cmd2_handler.get()
        if handler is not None:
            # Call whatever subcommand function was selected
            handler(ns)
        else:
            self._cmd.poutput('Stir pasta haphazardly')


class LoadableBadBase(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        super().__init__()
        self._dummy = dummy  # prevents autoload

    def do_cut(self, ns: argparse.Namespace) -> None:
        """Cut something"""
        handler = ns.cmd2_handler.get()
        if handler is not None:
            # Call whatever subcommand function was selected
            handler(ns)
        else:
            # No subcommand was provided, so call help
            self._cmd.poutput('This command does nothing without sub-parsers registered')
            self._cmd.do_help('cut')


@cmd2.with_default_category('Fruits')
class LoadableFruits(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        super().__init__()
        self._dummy = dummy  # prevents autoload

    def do_apple(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Apple')

    banana_parser = cmd2.Cmd2ArgumentParser()
    banana_parser.add_argument('direction', choices=['discs', 'lengthwise'])

    @cmd2.as_subcommand_to('cut', 'banana', banana_parser, help='Cut banana', aliases=['bananer'])
    def cut_banana(self, ns: argparse.Namespace) -> None:
        """Cut banana"""
        self._cmd.poutput('cutting banana: ' + ns.direction)


class LoadablePastaStir(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        super().__init__()
        self._dummy = dummy  # prevents autoload

    stir_pasta_vigor_parser = cmd2.Cmd2ArgumentParser()
    stir_pasta_vigor_parser.add_argument('frequency')

    @cmd2.as_subcommand_to('stir pasta', 'vigorously', stir_pasta_vigor_parser)
    def stir_pasta_vigorously(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput('stir the pasta vigorously')


@cmd2.with_default_category('Vegetables')
class LoadableVegetables(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        super().__init__()
        self._dummy = dummy  # prevents autoload

    def do_arugula(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Arugula')

    def complete_style_arg(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return ['quartered', 'diced']

    bokchoy_parser = cmd2.Cmd2ArgumentParser()
    bokchoy_parser.add_argument('style', completer=complete_style_arg)

    @cmd2.as_subcommand_to('cut', 'bokchoy', bokchoy_parser)
    def cut_bokchoy(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput('Bok Choy: ' + ns.style)


def test_subcommands(manual_command_sets_app) -> None:
    base_cmds = LoadableBase(1)
    badbase_cmds = LoadableBadBase(1)
    fruit_cmds = LoadableFruits(1)
    veg_cmds = LoadableVegetables(1)

    # installing subcommands without base command present raises exception
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.register_command_set(fruit_cmds)

    # if the base command is present but isn't an argparse command, expect exception
    manual_command_sets_app.register_command_set(badbase_cmds)
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.register_command_set(fruit_cmds)

    # verify that the commands weren't installed
    cmds_cats, cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()
    assert 'cut' in cmds_doc
    assert 'Fruits' not in cmds_cats

    # Now install the good base commands
    manual_command_sets_app.unregister_command_set(badbase_cmds)
    manual_command_sets_app.register_command_set(base_cmds)

    # verify that we catch an attempt to register subcommands when the commandset isn't installed
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._register_subcommands(fruit_cmds)

    cmd_result = manual_command_sets_app.app_cmd('cut')
    assert 'This command does nothing without sub-parsers registered' in cmd_result.stderr

    # verify that command set install without problems
    manual_command_sets_app.register_command_set(fruit_cmds)
    manual_command_sets_app.register_command_set(veg_cmds)
    cmds_cats, cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()
    assert 'Fruits' in cmds_cats

    text = ''
    line = f'cut {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)

    assert first_match is not None
    # check that the alias shows up correctly
    assert manual_command_sets_app.completion_matches == ['banana', 'bananer', 'bokchoy']

    cmd_result = manual_command_sets_app.app_cmd('cut banana discs')
    assert 'cutting banana: discs' in cmd_result.stdout

    text = ''
    line = f'cut bokchoy {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)

    assert first_match is not None
    # verify that argparse completer in commandset functions correctly
    assert manual_command_sets_app.completion_matches == ['diced', 'quartered']

    # verify that command set uninstalls without problems
    manual_command_sets_app.unregister_command_set(fruit_cmds)
    cmds_cats, cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()
    assert 'Fruits' not in cmds_cats

    # verify a double-unregister raises exception
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._unregister_subcommands(fruit_cmds)
    manual_command_sets_app.unregister_command_set(veg_cmds)

    # Disable command and verify subcommands still load and unload
    manual_command_sets_app.disable_command('cut', 'disabled for test')

    # verify that command set install without problems
    manual_command_sets_app.register_command_set(fruit_cmds)
    manual_command_sets_app.register_command_set(veg_cmds)

    manual_command_sets_app.enable_command('cut')

    cmds_cats, cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()
    assert 'Fruits' in cmds_cats

    text = ''
    line = f'cut {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)

    assert first_match is not None
    # check that the alias shows up correctly
    assert manual_command_sets_app.completion_matches == ['banana', 'bananer', 'bokchoy']

    text = ''
    line = f'cut bokchoy {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)

    assert first_match is not None
    # verify that argparse completer in commandset functions correctly
    assert manual_command_sets_app.completion_matches == ['diced', 'quartered']

    # disable again and verify can still uninstnall
    manual_command_sets_app.disable_command('cut', 'disabled for test')

    # verify that command set uninstalls without problems
    manual_command_sets_app.unregister_command_set(fruit_cmds)
    cmds_cats, cmds_doc, _cmds_undoc, _help_topics = manual_command_sets_app._build_command_info()
    assert 'Fruits' not in cmds_cats

    # verify a double-unregister raises exception
    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app._unregister_subcommands(fruit_cmds)

    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.unregister_command_set(base_cmds)

    manual_command_sets_app.unregister_command_set(veg_cmds)
    manual_command_sets_app.unregister_command_set(base_cmds)


def test_commandset_sigint(manual_command_sets_app) -> None:
    # shows that the command is able to continue execution if the sigint_handler
    # returns True that we've handled interrupting the command.
    class SigintHandledCommandSet(cmd2.CommandSet):
        def do_foo(self, _) -> None:
            self._cmd.poutput('in foo')
            self._cmd.sigint_handler(signal.SIGINT, None)
            self._cmd.poutput('end of foo')

        def sigint_handler(self) -> bool:
            return True

    cs1 = SigintHandledCommandSet()
    manual_command_sets_app.register_command_set(cs1)
    out = manual_command_sets_app.app_cmd('foo')
    assert 'in foo' in out.stdout
    assert 'end of foo' in out.stdout

    # shows that the command is interrupted if we don't report we've handled the sigint
    class SigintUnhandledCommandSet(cmd2.CommandSet):
        def do_bar(self, _) -> None:
            self._cmd.poutput('in do bar')
            self._cmd.sigint_handler(signal.SIGINT, None)
            self._cmd.poutput('end of do bar')

    cs2 = SigintUnhandledCommandSet()
    manual_command_sets_app.register_command_set(cs2)
    out = manual_command_sets_app.app_cmd('bar')
    assert 'in do bar' in out.stdout
    assert 'end of do bar' not in out.stdout


def test_nested_subcommands(manual_command_sets_app) -> None:
    base_cmds = LoadableBase(1)
    pasta_cmds = LoadablePastaStir(1)

    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.register_command_set(pasta_cmds)

    manual_command_sets_app.register_command_set(base_cmds)

    manual_command_sets_app.register_command_set(pasta_cmds)

    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.unregister_command_set(base_cmds)

    class BadNestedSubcommands(cmd2.CommandSet):
        def __init__(self, dummy) -> None:
            super().__init__()
            self._dummy = dummy  # prevents autoload

        stir_pasta_vigor_parser = cmd2.Cmd2ArgumentParser()
        stir_pasta_vigor_parser.add_argument('frequency')

        # stir sauce doesn't exist anywhere, this should fail
        @cmd2.as_subcommand_to('stir sauce', 'vigorously', stir_pasta_vigor_parser)
        def stir_pasta_vigorously(self, ns: argparse.Namespace) -> None:
            self._cmd.poutput('stir the pasta vigorously')

    with pytest.raises(CommandSetRegistrationError):
        manual_command_sets_app.register_command_set(BadNestedSubcommands(1))

    fruit_cmds = LoadableFruits(1)
    manual_command_sets_app.register_command_set(fruit_cmds)

    # validates custom namespace provider works correctly. Stir command will fail until
    # the cut command is called
    result = manual_command_sets_app.app_cmd('stir pasta vigorously everyminute')
    assert 'Need to cut before stirring' in result.stdout

    result = manual_command_sets_app.app_cmd('cut banana discs')
    assert 'cutting banana: discs' in result.stdout

    result = manual_command_sets_app.app_cmd('stir pasta vigorously everyminute')
    assert 'stir the pasta vigorously' in result.stdout


class AppWithSubCommands(cmd2.Cmd):
    """Class for testing usage of `as_subcommand_to` decorator directly in a Cmd2 subclass."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    cut_parser = cmd2.Cmd2ArgumentParser()
    cut_subparsers = cut_parser.add_subparsers(title='item', help='item to cut')

    @cmd2.with_argparser(cut_parser)
    def do_cut(self, ns: argparse.Namespace) -> None:
        """Cut something"""
        handler = ns.cmd2_handler.get()
        if handler is not None:
            # Call whatever subcommand function was selected
            handler(ns)
        else:
            # No subcommand was provided, so call help
            self.poutput('This command does nothing without sub-parsers registered')
            self.do_help('cut')

    banana_parser = cmd2.Cmd2ArgumentParser()
    banana_parser.add_argument('direction', choices=['discs', 'lengthwise'])

    @cmd2.as_subcommand_to('cut', 'banana', banana_parser, help='Cut banana', aliases=['bananer'])
    def cut_banana(self, ns: argparse.Namespace) -> None:
        """Cut banana"""
        self.poutput('cutting banana: ' + ns.direction)

    def complete_style_arg(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return ['quartered', 'diced']

    bokchoy_parser = cmd2.Cmd2ArgumentParser()
    bokchoy_parser.add_argument('style', completer=complete_style_arg)

    @cmd2.as_subcommand_to('cut', 'bokchoy', bokchoy_parser)
    def cut_bokchoy(self, _: argparse.Namespace) -> None:
        self.poutput('Bok Choy')


@pytest.fixture
def static_subcommands_app():
    return AppWithSubCommands(auto_load_commands=True)


def test_static_subcommands(static_subcommands_app) -> None:
    cmds_cats, _cmds_doc, _cmds_undoc, _help_topics = static_subcommands_app._build_command_info()
    assert 'Fruits' in cmds_cats

    text = ''
    line = f'cut {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, static_subcommands_app)

    assert first_match is not None
    # check that the alias shows up correctly
    assert static_subcommands_app.completion_matches == ['banana', 'bananer', 'bokchoy']

    text = ''
    line = f'cut bokchoy {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, static_subcommands_app)

    assert first_match is not None
    # verify that argparse completer in commandset functions correctly
    assert static_subcommands_app.completion_matches == ['diced', 'quartered']


complete_states_expected_self = None


@cmd2.with_default_category('With Completer')
class SupportFuncProvider(cmd2.CommandSet):
    """CommandSet which provides a support function (complete_states) to other CommandSets"""

    states = ('alabama', 'alaska', 'arizona', 'arkansas', 'california', 'colorado', 'connecticut', 'delaware')

    def __init__(self, dummy) -> None:
        """Dummy variable prevents this from being autoloaded in other tests"""
        super().__init__()

    def complete_states(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        assert self is complete_states_expected_self
        return self._cmd.basic_complete(text, line, begidx, endidx, self.states)


class SupportFuncUserSubclass1(SupportFuncProvider):
    """A sub-class of SupportFuncProvider which uses its support function"""

    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument('state', type=str, completer=SupportFuncProvider.complete_states)

    @cmd2.with_argparser(parser)
    def do_user_sub1(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput(f'something {ns.state}')


class SupportFuncUserSubclass2(SupportFuncProvider):
    """A second sub-class of SupportFuncProvider which uses its support function"""

    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument('state', type=str, completer=SupportFuncProvider.complete_states)

    @cmd2.with_argparser(parser)
    def do_user_sub2(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput(f'something {ns.state}')


class SupportFuncUserUnrelated(cmd2.CommandSet):
    """A CommandSet that isn't related to SupportFuncProvider which uses its support function"""

    def __init__(self, dummy) -> None:
        """Dummy variable prevents this from being autoloaded in other tests"""
        super().__init__()

    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument('state', type=str, completer=SupportFuncProvider.complete_states)

    @cmd2.with_argparser(parser)
    def do_user_unrelated(self, ns: argparse.Namespace) -> None:
        self._cmd.poutput(f'something {ns.state}')


def test_cross_commandset_completer(manual_command_sets_app, capsys) -> None:
    global complete_states_expected_self  # noqa: PLW0603
    # This tests the different ways to locate the matching CommandSet when completing an argparse argument.
    # Exercises the 3 cases in cmd2.Cmd._resolve_func_self() which is called during argparse tab completion.

    # Create all the CommandSets for these tests
    func_provider = SupportFuncProvider(1)
    user_sub1 = SupportFuncUserSubclass1(2)
    user_sub2 = SupportFuncUserSubclass2(3)
    user_unrelated = SupportFuncUserUnrelated(4)

    ####################################################################################################################
    # This exercises Case 1
    # If the CommandSet holding a command is a sub-class of the class that defines the completer function, then use that
    # CommandSet instance as self when calling the completer

    # Create instances of two different sub-class types to ensure no one removes the case 1 check in Cmd._resolve_func_self().
    # If that check is removed, testing with only 1 sub-class type will still pass. Testing it with two sub-class types
    # will fail and show that the case 1 check cannot be removed.
    manual_command_sets_app.register_command_set(user_sub1)
    manual_command_sets_app.register_command_set(user_sub2)

    text = ''
    line = f'user_sub1 {text}'
    endidx = len(line)
    begidx = endidx
    complete_states_expected_self = user_sub1
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)
    complete_states_expected_self = None

    assert first_match == 'alabama'
    assert manual_command_sets_app.completion_matches == list(SupportFuncProvider.states)

    assert (
        getattr(manual_command_sets_app.cmd_func('user_sub1').__func__, cmd2.constants.CMD_ATTR_HELP_CATEGORY)
        == 'With Completer'
    )

    manual_command_sets_app.unregister_command_set(user_sub2)
    manual_command_sets_app.unregister_command_set(user_sub1)

    ####################################################################################################################
    # This exercises Case 2
    # If the CommandSet holding a command is unrelated to the CommandSet holding the completer function, then search
    # all installed CommandSet instances for one that is an exact type match

    manual_command_sets_app.register_command_set(func_provider)
    manual_command_sets_app.register_command_set(user_unrelated)

    text = ''
    line = f'user_unrelated {text}'
    endidx = len(line)
    begidx = endidx
    complete_states_expected_self = func_provider
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)
    complete_states_expected_self = None

    assert first_match == 'alabama'
    assert manual_command_sets_app.completion_matches == list(SupportFuncProvider.states)

    manual_command_sets_app.unregister_command_set(user_unrelated)
    manual_command_sets_app.unregister_command_set(func_provider)

    ####################################################################################################################
    # This exercises Case 3
    # If the CommandSet holding a command is unrelated to the CommandSet holding the completer function,
    # and no exact type match can be found, but sub-class matches can be found and there is only a single
    # sub-class match, then use the lone sub-class match as the parent CommandSet.

    manual_command_sets_app.register_command_set(user_sub1)
    manual_command_sets_app.register_command_set(user_unrelated)

    text = ''
    line = f'user_unrelated {text}'
    endidx = len(line)
    begidx = endidx
    complete_states_expected_self = user_sub1
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)
    complete_states_expected_self = None

    assert first_match == 'alabama'
    assert manual_command_sets_app.completion_matches == list(SupportFuncProvider.states)

    manual_command_sets_app.unregister_command_set(user_unrelated)
    manual_command_sets_app.unregister_command_set(user_sub1)

    ####################################################################################################################
    # Error Case 1
    # If the CommandSet holding a command is unrelated to the CommandSet holding the completer function, then search
    # all installed CommandSet instances for one that is an exact type match, none are found
    # search for sub-class matches, also none are found.

    manual_command_sets_app.register_command_set(user_unrelated)

    text = ''
    line = f'user_unrelated {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)
    out, _err = capsys.readouterr()

    assert first_match is None
    assert manual_command_sets_app.completion_matches == []
    assert "Could not find CommandSet instance" in out

    manual_command_sets_app.unregister_command_set(user_unrelated)

    ####################################################################################################################
    # Error Case 2
    # If the CommandSet holding a command is unrelated to the CommandSet holding the completer function, then search
    # all installed CommandSet instances for one that is an exact type match, none are found
    # search for sub-class matches, more than 1 is found.

    manual_command_sets_app.register_command_set(user_sub1)
    manual_command_sets_app.register_command_set(user_sub2)
    manual_command_sets_app.register_command_set(user_unrelated)

    text = ''
    line = f'user_unrelated {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)
    out, _err = capsys.readouterr()

    assert first_match is None
    assert manual_command_sets_app.completion_matches == []
    assert "Could not find CommandSet instance" in out

    manual_command_sets_app.unregister_command_set(user_unrelated)
    manual_command_sets_app.unregister_command_set(user_sub2)
    manual_command_sets_app.unregister_command_set(user_sub1)


class CommandSetWithPathComplete(cmd2.CommandSet):
    def __init__(self, dummy) -> None:
        """Dummy variable prevents this from being autoloaded in other tests"""
        super().__init__()

    parser = cmd2.Cmd2ArgumentParser()
    parser.add_argument('path', nargs='+', help='paths', completer=cmd2.Cmd.path_complete)

    @cmd2.with_argparser(parser)
    def do_path(self, app: cmd2.Cmd, args) -> None:
        app.poutput(args.path)


def test_path_complete(manual_command_sets_app) -> None:
    test_set = CommandSetWithPathComplete(1)

    manual_command_sets_app.register_command_set(test_set)

    text = ''
    line = f'path {text}'
    endidx = len(line)
    begidx = endidx
    first_match = complete_tester(text, line, begidx, endidx, manual_command_sets_app)

    assert first_match is not None


def test_bad_subcommand() -> None:
    class BadSubcommandApp(cmd2.Cmd):
        """Class for testing usage of `as_subcommand_to` decorator directly in a Cmd2 subclass."""

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

        cut_parser = cmd2.Cmd2ArgumentParser()
        cut_subparsers = cut_parser.add_subparsers(title='item', help='item to cut')

        @cmd2.with_argparser(cut_parser)
        def do_cut(self, ns: argparse.Namespace) -> None:
            """Cut something"""

        banana_parser = cmd2.Cmd2ArgumentParser()
        banana_parser.add_argument('direction', choices=['discs', 'lengthwise'])

        @cmd2.as_subcommand_to('cut', 'bad name', banana_parser, help='This should fail')
        def cut_banana(self, ns: argparse.Namespace) -> None:
            """Cut banana"""
            self.poutput('cutting banana: ' + ns.direction)

    with pytest.raises(CommandSetRegistrationError):
        BadSubcommandApp()


def test_commandset_settables() -> None:
    # Define an arbitrary class with some attribute
    class Arbitrary:
        def __init__(self) -> None:
            self.some_value = 5

    # Declare a CommandSet with a settable of some arbitrary property
    class WithSettablesA(CommandSetBase):
        def __init__(self) -> None:
            super().__init__()

            self._arbitrary = Arbitrary()
            self._settable_prefix = 'addon'
            self.my_int = 11

            self.add_settable(
                Settable(
                    'arbitrary_value',
                    int,
                    'Some settable value',
                    settable_object=self._arbitrary,
                    settable_attrib_name='some_value',
                )
            )

    # Declare a CommandSet with an empty settable prefix
    class WithSettablesNoPrefix(CommandSetBase):
        def __init__(self) -> None:
            super().__init__()

            self._arbitrary = Arbitrary()
            self._settable_prefix = ''
            self.my_int = 11

            self.add_settable(
                Settable(
                    'another_value',
                    float,
                    'Some settable value',
                    settable_object=self._arbitrary,
                    settable_attrib_name='some_value',
                )
            )

    # Declare a commandset with duplicate settable name
    class WithSettablesB(CommandSetBase):
        def __init__(self) -> None:
            super().__init__()

            self._arbitrary = Arbitrary()
            self._settable_prefix = 'some'
            self.my_int = 11

            self.add_settable(
                Settable(
                    'arbitrary_value',
                    int,
                    'Some settable value',
                    settable_object=self._arbitrary,
                    settable_attrib_name='some_value',
                )
            )

    # create the command set and cmd2
    cmdset = WithSettablesA()
    arbitrary2 = Arbitrary()
    app = cmd2.Cmd(command_sets=[cmdset], auto_load_commands=False)
    app.str_value = ''
    app.add_settable(Settable('always_prefix_settables', bool, 'Prefix settables', app))
    app._settables['str_value'] = Settable('str_value', str, 'String value', app)

    assert 'arbitrary_value' in app.settables
    assert 'always_prefix_settables' in app.settables
    assert 'str_value' in app.settables

    # verify the settable shows up
    out, err = run_cmd(app, 'set')
    any('arbitrary_value' in line and '5' in line for line in out)

    out, err = run_cmd(app, 'set arbitrary_value')
    any('arbitrary_value' in line and '5' in line for line in out)

    # change the value and verify the value changed
    out, err = run_cmd(app, 'set arbitrary_value 10')
    expected = """
arbitrary_value - was: 5
now: 10
"""
    assert out == normalize(expected)
    out, err = run_cmd(app, 'set arbitrary_value')
    any('arbitrary_value' in line and '10' in line for line in out)

    # can't add to cmd2 now because commandset already has this settable
    with pytest.raises(KeyError):
        app.add_settable(Settable('arbitrary_value', int, 'This should fail', app))

    cmdset.add_settable(
        Settable('arbitrary_value', int, 'Replaced settable', settable_object=arbitrary2, settable_attrib_name='some_value')
    )

    # Can't add a settable to the commandset that already exists in cmd2
    with pytest.raises(KeyError):
        cmdset.add_settable(Settable('always_prefix_settables', int, 'This should also fail', cmdset))

    # Can't remove a settable from the CommandSet if it is elsewhere and not in the CommandSet
    with pytest.raises(KeyError):
        cmdset.remove_settable('always_prefix_settables')

    # verify registering a commandset with duplicate settable names fails
    cmdset_dupname = WithSettablesB()
    with pytest.raises(CommandSetRegistrationError):
        app.register_command_set(cmdset_dupname)

    # unregister the CommandSet and verify the settable is now gone
    app.unregister_command_set(cmdset)
    out, err = run_cmd(app, 'set')
    assert 'arbitrary_value' not in out
    out, err = run_cmd(app, 'set arbitrary_value')
    expected = """
Parameter 'arbitrary_value' not supported (type 'set' for list of parameters).
"""
    assert err == normalize(expected)

    # Add a commandset with no prefix
    cmdset_nopfx = WithSettablesNoPrefix()
    app.register_command_set(cmdset_nopfx)

    with pytest.raises(
        ValueError,
        match=r"Cannot force settable prefixes. CommandSet WithSettablesNoPrefix does not have a settable prefix defined.",
    ):
        app.always_prefix_settables = True

    app.unregister_command_set(cmdset_nopfx)

    # turn on prefixes and add the commandset back
    app.always_prefix_settables = True

    with pytest.raises(CommandSetRegistrationError):
        app.register_command_set(cmdset_nopfx)

    app.register_command_set(cmdset)

    # Verify the settable is back with the defined prefix.
    assert 'addon.arbitrary_value' in app.settables

    # rename the prefix and verify that the prefix changes everywhere
    cmdset._settable_prefix = 'some'
    assert 'addon.arbitrary_value' not in app.settables
    assert 'some.arbitrary_value' in app.settables

    out, err = run_cmd(app, 'set')
    any('some.arbitrary_value' in line and '5' in line for line in out)

    out, err = run_cmd(app, 'set some.arbitrary_value')
    any('some.arbitrary_value' in line and '5' in line for line in out)

    # verify registering a commandset with duplicate prefix and settable names fails
    with pytest.raises(CommandSetRegistrationError):
        app.register_command_set(cmdset_dupname)

    cmdset_dupname.remove_settable('arbitrary_value')

    app.register_command_set(cmdset_dupname)

    with pytest.raises(KeyError):
        cmdset_dupname.add_settable(
            Settable(
                'arbitrary_value',
                int,
                'Some settable value',
                settable_object=cmdset_dupname._arbitrary,
                settable_attrib_name='some_value',
            )
        )


class NsProviderSet(cmd2.CommandSet):
    # CommandSet which implements a namespace provider
    def __init__(self, dummy) -> None:
        # Use dummy argument so this won't be autoloaded by other tests
        super().__init__()

    def ns_provider(self) -> argparse.Namespace:
        ns = argparse.Namespace()
        # Save what was passed as self from with_argparser().
        ns.self = self
        return ns


class NsProviderApp(cmd2.Cmd):
    # Used to test namespace providers in CommandSets
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        super().__init__(*args, **kwargs)

    @cmd2.with_argparser(cmd2.Cmd2ArgumentParser(), ns_provider=NsProviderSet.ns_provider)
    def do_test_ns(self, args: argparse.Namespace) -> None:
        # Save args.self so the unit tests can read it.
        self.last_result = args.self


def test_ns_provider() -> None:
    """This exercises code in with_argparser() decorator that calls namespace providers"""
    ns_provider_set = NsProviderSet(1)
    app = NsProviderApp(auto_load_commands=False)

    # First test the case in which a namespace provider function resides in a CommandSet class which is registered.
    # with_argparser() will pass the CommandSet instance to the ns_provider() function.
    app.register_command_set(ns_provider_set)
    run_cmd(app, "test_ns")
    assert app.last_result == ns_provider_set

    # Now test the case in which a namespace provider function resides in a CommandSet class which is not registered.
    # with_argparser() will receive None from cmd2.Cmd._resolve_func_self() and therefore pass app as self to ns_provider().
    app.unregister_command_set(ns_provider_set)
    run_cmd(app, "test_ns")
    assert app.last_result == app



#!/usr/bin/env python
"""A simple example demonstrating the following how to emit a non-zero exit code in your cmd2 application."""

import cmd2


class ReplWithExitCode(cmd2.Cmd):
    """Example cmd2 application where we can specify an exit code when existing."""

    def __init__(self) -> None:
        super().__init__()

    @cmd2.with_argument_list
    def do_exit(self, arg_list: list[str]) -> bool:
        """Exit the application with an optional exit code.

        Usage:  exit [exit_code]
            Where:
                * exit_code - integer exit code to return to the shell
        """
        # If an argument was provided
        if arg_list:
            try:
                self.exit_code = int(arg_list[0])
            except ValueError:
                self.perror(f"{arg_list[0]} isn't a valid integer exit code")
                self.exit_code = 1

        return True


if __name__ == '__main__':
    import sys

    app = ReplWithExitCode()
    sys_exit_code = app.cmdloop()
    app.poutput(f'{sys.argv[0]!r} exiting with code: {sys_exit_code}')
    sys.exit(sys_exit_code)



#!/usr/bin/env python
"""This example demonstrates how to enable persistent readline history in your cmd2 application.

This will allow end users of your cmd2-based application to use the arrow keys and Ctrl+r in a manner which persists
across invocations of your cmd2 application.  This can make it much easier for them to use your application.
"""

import cmd2


class Cmd2PersistentHistory(cmd2.Cmd):
    """Basic example of how to enable persistent readline history within your cmd2 app."""

    def __init__(self, hist_file) -> None:
        """Configure the app to load persistent history from a file (both readline and cmd2 history command affected).

        :param hist_file: file to load history from at start and write it to at end
        """
        super().__init__(persistent_history_file=hist_file, persistent_history_length=500, allow_cli_args=False)
        self.prompt = 'ph> '

    # ... your class code here ...


if __name__ == '__main__':
    import sys

    history_file = '~/.persistent_history.cmd2'
    if len(sys.argv) > 1:
        history_file = sys.argv[1]

    app = Cmd2PersistentHistory(hist_file=history_file)
    sys.exit(app.cmdloop())



"""Example demonstrating that calling run_pyscript recursively inside another Python script isn't allowed"""

import os
import sys

app.cmd_echo = True
my_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
app('run_pyscript {}'.format(os.path.join(my_dir, 'stop.py')))



#!/usr/bin/env python3
"""Simple example demonstrating basic CommandSet usage."""

import cmd2
from cmd2 import (
    CommandSet,
    with_default_category,
)


@with_default_category('Default Category')
class MyBaseCommandSet(CommandSet):
    """Defines a default category for all sub-class CommandSets."""


class ChildInheritsParentCategories(MyBaseCommandSet):
    """This subclass doesn't declare any categories so all commands here are also categorized under 'Default Category'."""

    def do_hello(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Hello')

    def do_world(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('World')


@with_default_category('Non-Heritable Category', heritable=False)
class ChildOverridesParentCategoriesNonHeritable(MyBaseCommandSet):
    """This subclass overrides the 'Default Category' from the parent, but in a non-heritable fashion. Sub-classes of this
    CommandSet will not inherit this category and will, instead, inherit 'Default Category'.
    """

    def do_goodbye(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Goodbye')


class GrandchildInheritsGrandparentCategory(ChildOverridesParentCategoriesNonHeritable):
    """This subclass's parent class declared its default category non-heritable. Instead, it inherits the category defined
    by the grandparent class.
    """

    def do_aloha(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Aloha')


@with_default_category('Heritable Category')
class ChildOverridesParentCategories(MyBaseCommandSet):
    """This subclass is decorated with a default category that is heritable. This overrides the parent class's default
    category declaration.
    """

    def do_bonjour(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Bonjour')


class GrandchildInheritsHeritable(ChildOverridesParentCategories):
    """This subclass's parent declares a default category that overrides its parent. As a result, commands in this
    CommandSet will be categorized under 'Heritable Category'.
    """

    def do_monde(self, _: cmd2.Statement) -> None:
        self._cmd.poutput('Monde')


class ExampleApp(cmd2.Cmd):
    """Example to demonstrate heritable default categories."""

    def __init__(self) -> None:
        super().__init__()

    def do_something(self, _arg) -> None:
        self.poutput('this is the something command')


if __name__ == '__main__':
    app = ExampleApp()
    app.cmdloop()



import unittest
from unittest.mock import MagicMock, patch

import cmd2
import cmd2.cmd2  # to patch vt100_support
from cmd2 import rich_utils as ru


class TestAsyncAlert(unittest.TestCase):
    def test_async_alert_strips_ansi_when_allow_style_is_never(self):
        app = cmd2.Cmd()

        # Patch vt100_support to True
        with patch('cmd2.cmd2.vt100_support', True):
            # Patch threading functions
            mock_current_thread = MagicMock()
            mock_current_thread.name = "NotMainThread"

            with (
                patch('threading.current_thread', return_value=mock_current_thread),
                patch('threading.main_thread', return_value=MagicMock()),
                patch('cmd2.cmd2.rl_get_display_prompt', return_value='(Cmd) '),
                patch('cmd2.cmd2.readline.get_line_buffer', return_value=''),
                patch('cmd2.cmd2.rl_get_point', return_value=0),
                patch('cmd2.cmd2.rl_force_redisplay'),
                patch('sys.stdout', new_callable=MagicMock) as mock_stdout,
            ):
                # Set allow_style to NEVER
                app.allow_style = ru.AllowStyle.NEVER

                # Styled message
                msg = "\033[31mError\033[0m"

                # Call async_alert
                app.async_alert(msg)

                # Capture calls to write
                # mock_stdout.write.call_args_list -> [call(str), call(str)...]
                # We look at all written strings
                written_content = "".join([call.args[0] for call in mock_stdout.write.call_args_list])

                # Check that ANSI codes for color are NOT present
                if "\033[31m" in written_content:
                    raise AssertionError(f"Found ANSI color code in output: {written_content!r}")
                if "Error" not in written_content:
                    raise AssertionError(f"Message 'Error' not found in output: {written_content!r}")


if __name__ == '__main__':
    unittest.main()



"""Cmd2 functional testing based on transcript"""

import os
import random
import re
import sys
import tempfile
from typing import NoReturn
from unittest import (
    mock,
)

import pytest

import cmd2
from cmd2 import (
    transcript,
)
from cmd2.utils import (
    Settable,
    StdSim,
)

from .conftest import (
    run_cmd,
    verify_help_text,
)


class CmdLineApp(cmd2.Cmd):
    MUMBLES = ('like', '...', 'um', 'er', 'hmmm', 'ahh')
    MUMBLE_FIRST = ('so', 'like', 'well')
    MUMBLE_LAST = ('right?',)

    def __init__(self, *args, **kwargs) -> None:
        self.maxrepeats = 3

        super().__init__(*args, multiline_commands=['orate'], **kwargs)

        # Make maxrepeats settable at runtime
        self.add_settable(Settable('maxrepeats', int, 'Max number of `--repeat`s allowed', self))

        self.intro = 'This is an intro banner ...'

    speak_parser = cmd2.Cmd2ArgumentParser()
    speak_parser.add_argument('-p', '--piglatin', action="store_true", help="atinLay")
    speak_parser.add_argument('-s', '--shout', action="store_true", help="N00B EMULATION MODE")
    speak_parser.add_argument('-r', '--repeat', type=int, help="output [n] times")

    @cmd2.with_argparser(speak_parser, with_unknown_args=True)
    def do_speak(self, opts, arg) -> None:
        """Repeats what you tell me to."""
        arg = ' '.join(arg)
        if opts.piglatin:
            arg = f'{arg[1:]}{arg[0]}ay'
        if opts.shout:
            arg = arg.upper()
        repetitions = opts.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            self.poutput(arg)
            # recommend using the poutput function instead of
            # self.stdout.write or "print", because Cmd allows the user
            # to redirect output

    do_say = do_speak  # now "say" is a synonym for "speak"
    do_orate = do_speak  # another synonym, but this one takes multi-line input

    mumble_parser = cmd2.Cmd2ArgumentParser()
    mumble_parser.add_argument('-r', '--repeat', type=int, help="output [n] times")

    @cmd2.with_argparser(mumble_parser, with_unknown_args=True)
    def do_mumble(self, opts, arg) -> None:
        """Mumbles what you tell me to."""
        repetitions = opts.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            output = []
            if random.random() < 0.33:
                output.append(random.choice(self.MUMBLE_FIRST))
            for word in arg:
                if random.random() < 0.40:
                    output.append(random.choice(self.MUMBLES))
                output.append(word)
            if random.random() < 0.25:
                output.append(random.choice(self.MUMBLE_LAST))
            self.poutput(' '.join(output))

    def do_nothing(self, statement) -> None:
        """Do nothing and output nothing"""

    def do_keyboard_interrupt(self, _) -> NoReturn:
        raise KeyboardInterrupt('Interrupting this command')


def test_commands_at_invocation() -> None:
    testargs = ["prog", "say hello", "say Gracie", "quit"]
    expected = "This is an intro banner ...\nhello\nGracie\n"
    with mock.patch.object(sys, 'argv', testargs):
        app = CmdLineApp()

    app.stdout = StdSim(app.stdout)
    app.cmdloop()
    out = app.stdout.getvalue()
    assert out == expected


@pytest.mark.parametrize(
    ('filename', 'feedback_to_output'),
    [
        ('bol_eol.txt', False),
        ('characterclass.txt', False),
        ('dotstar.txt', False),
        ('extension_notation.txt', False),
        ('from_cmdloop.txt', True),
        ('multiline_no_regex.txt', False),
        ('multiline_regex.txt', False),
        ('no_output.txt', False),
        ('no_output_last.txt', False),
        ('singleslash.txt', False),
        ('slashes_escaped.txt', False),
        ('slashslash.txt', False),
        ('spaces.txt', False),
        ('word_boundaries.txt', False),
    ],
)
def test_transcript(request, capsys, filename, feedback_to_output) -> None:
    # Get location of the transcript
    test_dir = os.path.dirname(request.module.__file__)
    transcript_file = os.path.join(test_dir, 'transcripts', filename)

    # Need to patch sys.argv so cmd2 doesn't think it was called with
    # arguments equal to the py.test args
    testargs = ['prog', '-t', transcript_file]
    with mock.patch.object(sys, 'argv', testargs):
        # Create a cmd2.Cmd() instance and make sure basic settings are
        # like we want for test
        app = CmdLineApp()

    app.feedback_to_output = feedback_to_output

    # Run the command loop
    sys_exit_code = app.cmdloop()
    assert sys_exit_code == 0

    # Check for the unittest "OK" condition for the 1 test which ran
    expected_start = ".\n----------------------------------------------------------------------\nRan 1 test in"
    expected_end = "s\n\nOK\n"
    _, err = capsys.readouterr()
    assert err.startswith(expected_start)
    assert err.endswith(expected_end)


def test_history_transcript() -> None:
    app = CmdLineApp()
    app.stdout = StdSim(app.stdout)
    run_cmd(app, 'orate this is\na /multiline/\ncommand;\n')
    run_cmd(app, 'speak /tmp/file.txt is not a regex')

    expected = r"""(Cmd) orate this is
> a /multiline/
> command;
this is a \/multiline\/ command
(Cmd) speak /tmp/file.txt is not a regex
\/tmp\/file.txt is not a regex
"""

    # make a tmp file
    fd, history_fname = tempfile.mkstemp(prefix='', suffix='.txt')
    os.close(fd)

    # tell the history command to create a transcript
    run_cmd(app, f'history -t "{history_fname}"')

    # read in the transcript created by the history command
    with open(history_fname) as f:
        xscript = f.read()

    assert xscript == expected


def test_history_transcript_bad_path(mocker) -> None:
    app = CmdLineApp()
    app.stdout = StdSim(app.stdout)
    run_cmd(app, 'orate this is\na /multiline/\ncommand;\n')
    run_cmd(app, 'speak /tmp/file.txt is not a regex')

    # Bad directory
    history_fname = '~/fakedir/this_does_not_exist.txt'
    _out, err = run_cmd(app, f'history -t "{history_fname}"')
    assert "is not a directory" in err[0]

    # Cause os.open to fail and make sure error gets printed
    mock_remove = mocker.patch('builtins.open')
    mock_remove.side_effect = OSError

    history_fname = 'outfile.txt'
    _out, err = run_cmd(app, f'history -t "{history_fname}"')
    assert "Error saving transcript file" in err[0]


def test_run_script_record_transcript(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    filename = os.path.join(test_dir, 'scripts', 'help.txt')

    assert base_app._script_dir == []
    assert base_app._current_script_dir is None

    # make a tmp file to use as a transcript
    fd, transcript_fname = tempfile.mkstemp(prefix='', suffix='.trn')
    os.close(fd)

    # Execute the run_script command with the -t option to generate a transcript
    run_cmd(base_app, f'run_script {filename} -t {transcript_fname}')

    assert base_app._script_dir == []
    assert base_app._current_script_dir is None

    # read in the transcript created by the history command
    with open(transcript_fname) as f:
        xscript = f.read()

    assert xscript.startswith('(Cmd) help -v\n')
    verify_help_text(base_app, xscript)


def test_generate_transcript_stop(capsys) -> None:
    # Verify transcript generation stops when a command returns True for stop
    app = CmdLineApp()

    # Make a tmp file to use as a transcript
    fd, transcript_fname = tempfile.mkstemp(prefix='', suffix='.trn')
    os.close(fd)

    # This should run all commands
    commands = ['help', 'set']
    app._generate_transcript(commands, transcript_fname)
    _, err = capsys.readouterr()
    assert err.startswith("2 commands")

    # Since quit returns True for stop, only the first 2 commands will run
    commands = ['help', 'quit', 'set']
    app._generate_transcript(commands, transcript_fname)
    _, err = capsys.readouterr()
    assert err.startswith("Command 2 triggered a stop")

    # keyboard_interrupt command should stop the loop and not run the third command
    commands = ['help', 'keyboard_interrupt', 'set']
    app._generate_transcript(commands, transcript_fname)
    _, err = capsys.readouterr()
    assert err.startswith("Interrupting this command\nCommand 2 triggered a stop")


@pytest.mark.parametrize(
    ('expected', 'transformed'),
    [
        # strings with zero or one slash or with escaped slashes means no regular
        # expression present, so the result should just be what re.escape returns.
        # we don't use static strings in these tests because re.escape behaves
        # differently in python 3.7+ than in prior versions
        ('text with no slashes', re.escape('text with no slashes')),
        ('specials .*', re.escape('specials .*')),
        ('use 2/3 cup', re.escape('use 2/3 cup')),
        ('/tmp is nice', re.escape('/tmp is nice')),
        ('slash at end/', re.escape('slash at end/')),
        # escaped slashes
        (r'not this slash\/ or this one\/', re.escape('not this slash/ or this one/')),
        # regexes
        ('/.*/', '.*'),
        ('specials ^ and + /[0-9]+/', re.escape('specials ^ and + ') + '[0-9]+'),
        (r'/a{6}/ but not \/a{6} with /.*?/ more', 'a{6}' + re.escape(' but not /a{6} with ') + '.*?' + re.escape(' more')),
        (r'not \/, use /\|?/, not \/', re.escape('not /, use ') + r'\|?' + re.escape(', not /')),
        # inception: slashes in our regex. backslashed on input, bare on output
        (r'not \/, use /\/?/, not \/', re.escape('not /, use ') + '/?' + re.escape(', not /')),
        (r'lots /\/?/ more /.*/ stuff', re.escape('lots ') + '/?' + re.escape(' more ') + '.*' + re.escape(' stuff')),
    ],
)
def test_parse_transcript_expected(expected, transformed) -> None:
    app = CmdLineApp()

    class TestMyAppCase(transcript.Cmd2TestCase):
        cmdapp = app

    testcase = TestMyAppCase()
    assert testcase._transform_transcript_expected(expected) == transformed


def test_transcript_failure(request, capsys) -> None:
    # Get location of the transcript
    test_dir = os.path.dirname(request.module.__file__)
    transcript_file = os.path.join(test_dir, 'transcripts', 'failure.txt')

    # Need to patch sys.argv so cmd2 doesn't think it was called with
    # arguments equal to the py.test args
    testargs = ['prog', '-t', transcript_file]
    with mock.patch.object(sys, 'argv', testargs):
        # Create a cmd2.Cmd() instance and make sure basic settings are
        # like we want for test
        app = CmdLineApp()

    app.feedback_to_output = False

    # Run the command loop
    sys_exit_code = app.cmdloop()
    assert sys_exit_code != 0

    expected_start = "File "
    expected_end = "s\n\nFAILED (failures=1)\n\n"
    _, err = capsys.readouterr()
    assert err.startswith(expected_start)
    assert err.endswith(expected_end)


def test_transcript_no_file(request, capsys) -> None:
    # Need to patch sys.argv so cmd2 doesn't think it was called with
    # arguments equal to the py.test args
    testargs = ['prog', '-t']
    with mock.patch.object(sys, 'argv', testargs):
        app = CmdLineApp()

    app.feedback_to_output = False

    # Run the command loop
    sys_exit_code = app.cmdloop()
    assert sys_exit_code != 0

    # Check for the unittest "OK" condition for the 1 test which ran
    expected = 'No test files found - nothing to test\n'
    _, err = capsys.readouterr()
    assert err == expected



#!/usr/bin/env python
"""A complex example demonstrating a variety of methods to load CommandSets using a mix of command decorators.

Includes examples of how to integrate tab completion with argparse-based commands.
"""

import argparse
from collections.abc import Iterable

from modular_commands.commandset_basic import (
    BasicCompletionCommandSet,
)
from modular_commands.commandset_complex import (
    CommandSetA,
)
from modular_commands.commandset_custominit import (
    CustomInitCommandSet,
)

from cmd2 import (
    Cmd,
    Cmd2ArgumentParser,
    CommandSet,
    with_argparser,
)


class WithCommandSets(Cmd):
    def __init__(self, command_sets: Iterable[CommandSet] | None = None) -> None:
        """Cmd2 application to demonstrate a variety of methods for loading CommandSets."""
        super().__init__(command_sets=command_sets)
        self.sport_item_strs = ['Bat', 'Basket', 'Basketball', 'Football', 'Space Ball']

    def choices_provider(self) -> list[str]:
        """A choices provider is useful when the choice list is based on instance data of your application."""
        return self.sport_item_strs

    # Parser for example command
    example_parser = Cmd2ArgumentParser(
        description="Command demonstrating tab completion with argparse\nNotice even the flags of this command tab complete"
    )

    # Tab complete from a list using argparse choices. Set metavar if you don't
    # want the entire choices list showing in the usage text for this command.
    example_parser.add_argument(
        '--choices', choices=['some', 'choices', 'here'], metavar="CHOICE", help="tab complete using choices"
    )

    # Tab complete from choices provided by a choices provider
    example_parser.add_argument(
        '--choices_provider', choices_provider=choices_provider, help="tab complete using a choices_provider"
    )

    # Tab complete using a completer
    example_parser.add_argument('--completer', completer=Cmd.path_complete, help="tab complete using a completer")

    @with_argparser(example_parser)
    def do_example(self, _: argparse.Namespace) -> None:
        """An example command."""
        self.poutput("I do nothing")


if __name__ == '__main__':
    import sys

    print("Starting")
    my_sets = [BasicCompletionCommandSet(), CommandSetA(), CustomInitCommandSet('First argument', 'Second argument')]
    app = WithCommandSets(command_sets=my_sets)
    sys.exit(app.cmdloop())



#!/usr/bin/env python
"""A sample cmd application that shows how to trivially migrate a cmd application to use cmd2."""

# import cmd2 as cmd  # noqa: ERA001
import cmd  # Comment this line and uncomment the one above to migrate to cmd2
import random


class CmdLineApp(cmd.Cmd):
    """Example cmd application."""

    MUMBLES = ('like', '...', 'um', 'er', 'hmmm', 'ahh')
    MUMBLE_FIRST = ('so', 'like', 'well')
    MUMBLE_LAST = ('right?',)

    def do_exit(self, _line) -> bool:
        """Exit the application."""
        return True

    do_EOF = do_exit  # noqa: N815
    do_quit = do_exit

    def do_speak(self, line) -> None:
        """Repeats what you tell me to."""
        print(line, file=self.stdout)

    do_say = do_speak

    def do_mumble(self, line) -> None:
        """Mumbles what you tell me to."""
        words = line.split(' ')
        output = []
        if random.random() < 0.33:
            output.append(random.choice(self.MUMBLE_FIRST))
        for word in words:
            if random.random() < 0.40:
                output.append(random.choice(self.MUMBLES))
            output.append(word)
        if random.random() < 0.25:
            output.append(random.choice(self.MUMBLE_LAST))
        print(' '.join(output), file=self.stdout)


if __name__ == '__main__':
    import sys

    c = CmdLineApp()
    sys.exit(c.cmdloop())



#!/usr/bin/env python
import os
import sys

print(f"Running Python script {os.path.basename(sys.argv[0])!r} which was called with {len(sys.argv) - 1} arguments")
for i, arg in enumerate(sys.argv[1:]):
    print(f"arg {i + 1}: {arg!r}")



#!/usr/bin/env python
"""A simple example demonstrating how to remove unused commands.

Commands can be removed from  help menu and tab completion by appending their command name to the hidden_commands list.
These commands will still exist and can be executed and help can be retrieved for them by
name, they just won't clutter the help menu.

Commands can also be removed entirely by using Python's "del".
"""

import cmd2


class RemoveBuiltinCommands(cmd2.Cmd):
    """Example cmd2 application where we remove some unused built-in commands."""

    def __init__(self) -> None:
        super().__init__()

        # To hide commands from displaying in the help menu, add them to the hidden_commands list
        self.hidden_commands.append('history')

        # To remove built-in commands entirely, delete their "do_*" function from the cmd2.Cmd class
        del cmd2.Cmd.do_edit


if __name__ == '__main__':
    import sys

    app = RemoveBuiltinCommands()
    sys.exit(app.cmdloop())



# This script demonstrates that cmd2 can capture sys.stdout and self.stdout when both point to the same stream.
# Set base_app.self_in_py to True before running this script.
print("print")
self.poutput("poutput")
